import re
import copy

# https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts

# stdlib
import os
from inspect import isclass
from lxml import etree
import json

# local
from jnpr.junos.exception import RpcError
from jnpr.junos.utils.start_shell import StartShell
from jnpr.junos.factory.state_machine import StateMachine
from jnpr.junos.factory.to_json import TableJSONEncoder

from jinja2 import Template

HAS_NTC_TEMPLATE = False
try:
    from ntc_templates import parse as ntc_parse

    HAS_NTC_TEMPLATE = True
except:
    pass

import logging

logger = logging.getLogger("jnpr.junos.factory.cmdtable")


class CMDTable(object):
    def __init__(self, dev=None, raw=None, path=None, template_dir=None):
        """
        :dev: Device instance
        :raw: string blob of the command output
        :path: file path to XML, to be used rather than :dev:
        :template_dir: To look for textfsm templates in this folder first
        """
        self._dev = dev
        self.xml = None
        self.view = None
        self.ITEM_FILTER = "name"
        self._key_list = []
        self._path = path
        self._parser = None
        self.output = None
        self.data = raw
        self.template_dir = template_dir

    # -------------------------------------------------------------------------
    # PUBLIC METHODS
    # -------------------------------------------------------------------------

    def get(self, *vargs, **kvargs):
        r"""
        Retrieve the XML (string blob under <output> tag of table data from the
        Device instance and returns back the Table instance - for call-chaining
        purposes.

        If the Table was created with a :path: rather than a Device,
        then this method will load the string blob from that file.  In this
        case, the \*vargs, and \**kvargs are not used.

        ALIAS: __call__

        :vargs:
          [0] is the table :arg_key: value.  This is used so that
          the caller can retrieve just one item from the table without
          having to know the Junos RPC argument.

        :kvargs:
          these are the name/value pairs relating to the specific Junos
          command attached to the table.  For example, if the command
          is 'show ithrottle id {{ id }}', there is a parameter 'id'.
          Any valid command argument can be passed to :kvargs: to further filter
          the results of the :get(): operation.  neato!

        NOTES:
          If you need to create a 'stub' for unit-testing
          purposes, you want to create a subclass of your table and
          overload this methods.
        """
        self._clearkeys()

        if self._path and self.data:
            raise AttributeError("path and data are mutually exclusive")
        if self._path is not None:
            # for loading from local file-path
            with open(self._path, "r") as fp:
                self.data = fp.read().strip()
            if self.data.startswith("<output>") and self.data.endswith("</output>"):
                self.data = etree.fromstring(self.data).text

        if "target" in kvargs:
            self.TARGET = kvargs["target"]

        if "key" in kvargs:
            self.KEY = kvargs["key"]

        if "key_items" in kvargs:
            self.KEY_ITEMS = kvargs["key_items"]

        if "filters" in kvargs:
            self.VIEW.FILTERS = (
                [kvargs["filters"]]
                if isinstance(kvargs["filters"], str)
                else kvargs["filters"]
            )

        cmd_args = self.CMD_ARGS.copy()
        if "args" in kvargs and isinstance(kvargs["args"], dict):
            cmd_args.update(kvargs["args"])

        if len(cmd_args) > 0:
            self.GET_CMD = Template(self.GET_CMD).render(**cmd_args)

        if self.data is None:
            # execute the Junos RPC to retrieve the table
            if hasattr(self, "TARGET"):
                if self.TARGET is None:
                    raise ValueError('"target" value not provided')
                rpc_args = {
                    "target": self.TARGET,
                    "command": self.GET_CMD,
                    "timeout": "0",
                }
                try:
                    self.xml = getattr(self.RPC, "request_pfe_execute")(**rpc_args)
                    self.data = self.xml.text
                    ver_info = self._dev.facts.get("version_info")
                    if ver_info and ver_info.major[0] <= 15:
                        # Junos <=15.x output has output something like below
                        #
                        # <rpc-reply>
                        # <output>
                        # SENT: Ukern command: show memory
                        # GOT:
                        # GOT: ID      Base      Total(b)       Free(b)     Used(b)
                        # GOT: --  --------     ---------     ---------   ---------
                        # GOT:  0  44e72078    1882774284    1689527364   193246920
                        # GOT:  1  b51ffb88      67108860      57651900     9456960
                        # GOT:  2  bcdfffe0      52428784      52428784           0
                        # GOT:  3  b91ffb88      62914556      62914556           0
                        # LOCAL: End of file
                        # </output>
                        # </rpc-reply>
                        # hence need to do cleanup
                        self.data = self.data.replace("GOT: ", "")
                except RpcError:
                    with StartShell(self.D) as ss:
                        ret = ss.run(
                            'cprod -A %s -c "%s"' % (self.TARGET, self.GET_CMD)
                        )
                        if ret[0]:
                            self.data = ret[1]
            else:
                self.xml = self.RPC.cli(self.GET_CMD)
                self.data = self.xml.text

        if self.USE_TEXTFSM:
            if HAS_NTC_TEMPLATE:
                self.output = self._parse_textfsm(
                    platform=self.PLATFORM, command=self.GET_CMD, raw=self.data
                )
            else:
                raise ImportError(
                    "ntc_template is missing. Need to be installed explicitly."
                )
        else:
            # state machine
            sm = StateMachine(self)
            self.output = sm.parse(self.data.splitlines())

        # returning self for call-chaining purposes, yo!
        return self

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def D(self):
        """the Device instance"""
        return self._dev

    @property
    def CLI(self):
        """the Device.cli instance"""
        return self.D.cli

    @property
    def RPC(self):
        """the Device.rpc instance"""
        return self.D.rpc

    @property
    def view(self):
        """returns the current view assigned to this table"""
        return self._view

    @view.setter
    def view(self, cls):
        """assigns a new view to the table"""
        if cls is None:
            self._view = None
            return

        if not isclass(cls):
            raise ValueError("Must be given RunstatView class")

        self._view = cls

    @property
    def hostname(self):
        return self.D.hostname

    @property
    def key_list(self):
        """the list of keys, as property for caching"""
        return self._key_list

    # -------------------------------------------------------------------------
    # PRIVATE METHODS
    # -------------------------------------------------------------------------

    def _assert_data(self):
        if self.data is None:
            raise RuntimeError("Table is empty, use get()")

    def _clearkeys(self):
        self._key_list = []

    # -------------------------------------------------------------------------
    # PUBLIC METHODS
    # -------------------------------------------------------------------------

    # ------------------------------------------------------------------------
    # keys
    # ------------------------------------------------------------------------

    def _keys(self):
        """return a list of data item keys from the data string"""

        return self.output.keys()

    def keys(self):
        # if the key_list has been cached, then use it
        if len(self.key_list):
            return self.key_list

        # otherwise, build the list of keys into the cache
        self._key_list = self._keys()
        return self._key_list

    # ------------------------------------------------------------------------
    # values
    # ------------------------------------------------------------------------

    def values(self):
        """returns list of table entry items()"""

        self._assert_data()
        return self.output.values()

    # ------------------------------------------------------------------------
    # items
    # ------------------------------------------------------------------------

    def items(self):
        """returns list of tuple(name,values) for each table entry"""
        return self.output.items()

    def to_json(self):
        """
        :returns: JSON encoded string of entire Table contents
        """
        return json.dumps(self, cls=TableJSONEncoder)

    # -------------------------------------------------------------------------
    # OVERLOADS
    # -------------------------------------------------------------------------

    __call__ = get

    def __repr__(self):
        cls_name = self.__class__.__name__
        source = self.D.hostname if self.D is not None else self._path

        if self.data is None:
            return "%s:%s - Table empty" % (cls_name, source)
        else:
            n_items = len(self.keys())
            return "%s:%s: %s items" % (cls_name, source, n_items)

    def __len__(self):
        self._assert_data()
        return len(self.keys())

    def __iter__(self):
        """iterate over each time in the table"""
        self._assert_data()

        for key, value in self.output.items():
            yield key, value

    def __getitem__(self, value):
        """
        returns a table item. If a table view is set (should be by default)
        then the item will be converted to the view upon return.  if there is
        no table view, then the XML object will be returned.

        :value:
          for <string>, this will perform a select based on key-name
          for <tuple>, this will perform a select based on compsite key-name
          for <int>, this will perform a select based by position, like <list>
            [0] is the first item
            [-1] is the last item
          when it is a <slice> then this will return a <list> of View widgets
        """
        self._assert_data()
        return self.output[value]

    def __contains__(self, key):
        """membership for use with 'in'"""
        return bool(key in self.keys())

    # ------------------------------------------------------------------------
    # textfsm
    # ------------------------------------------------------------------------

    def _parse_textfsm(self, platform=None, command=None, raw=None):
        """
        textfsm returns list of dict, make it JSON/dict

        :param platform: vendor platform, for ex cisco_xr
        :param command: cli command to be parsed
        :param raw: string blob output from the cli command execution
        :return: dict of parsed data.
        """
        command = command.replace("/", "-")

        attrs = dict(Command=command, Platform=platform)

        template = None
        template_dir = None
        if self.template_dir is not None:
            # we dont need index file for lookup
            index = None
            template_path = os.path.join(
                self.template_dir,
                "{}_{}.textfsm".format(platform, "_".join(command.split())),
            )
            if not os.path.exists(template_path):
                msg = "Template file %s missing" % template_path
                logger.error(msg)
                raise FileNotFoundError(msg)
            else:
                template = template_path
                template_dir = self.template_dir
        if template_dir is None:
            index = "index"
            template_dir = ntc_parse._get_template_dir()

        cli_table = ntc_parse.clitable.CliTable(index, template_dir)
        try:
            cli_table.ParseCmd(raw, attrs, template)
        except ntc_parse.clitable.CliTableError as ex:
            logger.error(
                'Unable to parse command "%s" on platform %s' % (command, platform)
            )
            raise ex
        return self._filter_output(cli_table)

    def _filter_output(self, cli_table):
        """
        textfsm return list of list, covert it into more consumable format

        :param cli_table: CLiTable object from textfsm
        :return: dict of key, fields and its values, list of dict when key is None
        """
        self._set_key(cli_table)

        fields = self.VIEW.FIELDS if self.VIEW is not None else {}
        reverse_fields = {val: key for key, val in fields.items()}
        if self.KEY is None:
            cli_table_size = cli_table.size
            if cli_table_size > 1:
                raise KeyError(
                    "Key is Mandatory for parsed o/p of %s " "length" % cli_table_size
                )
            elif cli_table_size == 1:
                temp_dict = self._parse_row(cli_table[1], cli_table, reverse_fields)
                logger.debug("For Null Key, data returned: {}".format(temp_dict))
                return temp_dict
        output = {}
        for row in cli_table:
            temp_dict = self._parse_row(row, cli_table, reverse_fields)
            logger.debug("data at index {} is {}".format(row.row, temp_dict))
            if isinstance(self.KEY, list):
                key_list = []
                for key in self.KEY:
                    if key not in fields:
                        key_list.append(temp_dict.pop(key))
                    else:
                        key_list.append(temp_dict[key])
                output[tuple(key_list)] = temp_dict
            elif self.KEY in temp_dict:
                if self.KEY not in fields:
                    output[temp_dict.pop(self.KEY)] = temp_dict
                else:
                    output[temp_dict[self.KEY]] = temp_dict
            else:
                logger.debug("Key {} not present in {}".format(self.KEY, temp_dict))
        return output

    def _parse_row(self, row, cli_table, reverse_fields):
        temp_dict = {}
        for index, element in enumerate(row):
            key = cli_table.header[index]
            if self.KEY and key in self.KEY:
                temp_dict[key] = element
            if reverse_fields:
                if key in reverse_fields:
                    temp_dict[reverse_fields[key]] = element
            else:
                temp_dict[key] = element
        return temp_dict

    def _set_key(self, cli_table):
        """
        Preference to user provided key, then template and at last default
        Checks and update if we need KEY from template file

        :param cli_table: CLiTable object from textfsm
        :return:
        """
        if self.KEY == "name" and len(cli_table._keys) > 0:
            template_keys = list(cli_table._keys)
            self.KEY = template_keys[0] if len(template_keys) == 1 else template_keys
        logger.debug("KEY being used: {}".format(self.KEY))



from copy import deepcopy
import logging

# 3rd-party
from lxml import etree
from lxml.builder import E

# local
from jnpr.junos.factory.table import Table
from jnpr.junos.jxml import remove_namespaces, remove_namespaces_and_spaces
from jnpr.junos.decorators import checkSAXParserDecorator

logger = logging.getLogger("jnpr.junos.factory.optable")


class OpTable(Table):
    # -------------------------------------------------------------------------
    # PUBLIC METHODS
    # -------------------------------------------------------------------------

    @checkSAXParserDecorator
    def get(self, *vargs, **kvargs):
        r"""
        Retrieve the XML table data from the Device instance and
        returns back the Table instance - for call-chaining purposes.

        If the Table was created with a :path: rather than a Device,
        then this method will load the XML from that file.  In this
        case, the \*vargs, and \**kvargs are not used.

        ALIAS: __call__

        :vargs:
          [0] is the table :arg_key: value.  This is used so that
          the caller can retrieve just one item from the table without
          having to know the Junos RPC argument.

        :kvargs:
          these are the name/value pairs relating to the specific Junos
          XML command attached to the table.  For example, if the RPC
          is 'get-route-information', there are parameters such as
          'table' and 'destination'.  Any valid RPC argument can be
          passed to :kvargs: to further filter the results of the :get():
          operation.  neato!

        NOTES:
          If you need to create a 'stub' for unit-testing
          purposes, you want to create a subclass of your table and
          overload this methods.
        """
        self._clearkeys()

        if self._path is not None:
            # for loading from local file-path
            self.xml = remove_namespaces(etree.parse(self._path).getroot())
            return self

        if self._lxml is not None:
            return self

        argkey = vargs[0] if len(vargs) else None

        rpc_args = {}

        if self._use_filter:
            try:
                filter_xml = generate_sax_parser_input(self)
                rpc_args["filter_xml"] = filter_xml
            except Exception as ex:
                logger.debug("Not able to create SAX parser input due to " "'%s'" % ex)
            self.D.transform = lambda: remove_namespaces_and_spaces

        self.D.transform = lambda: remove_namespaces_and_spaces
        rpc_args.update(self.GET_ARGS)  # copy default args
        # saltstack get_table pass args as named keyword
        if "args" in kvargs and isinstance(kvargs["args"], dict):
            rpc_args.update(kvargs.pop("args"))
        rpc_args.update(kvargs)  # copy caller provided args

        if hasattr(self, "GET_KEY") and argkey is not None:
            rpc_args.update({self.GET_KEY: argkey})

        # execute the Junos RPC to retrieve the table
        self.xml = getattr(self.RPC, self.GET_RPC)(**rpc_args)

        # returning self for call-chaining purposes, yo!
        return self


def generate_sax_parser_input(obj):
    """
    Used to generate xml object from Table/view to be used in SAX parsing
    Args:
    obj: self object which contains table/view details

    Returns: lxml etree object to be used as sax parser input

    """
    item_tags = []
    if "/" in obj.ITEM_XPATH:
        item_tags = obj.ITEM_XPATH.split("/")
        parser_ingest = E(item_tags.pop(-1), E(obj.ITEM_NAME_XPATH))
    else:
        parser_ingest = E(obj.ITEM_XPATH, E(obj.ITEM_NAME_XPATH))
    local_field_dict = deepcopy(obj.VIEW.FIELDS)
    # first make element out of group fields
    if obj.VIEW.GROUPS:
        for group, group_xpath in obj.VIEW.GROUPS.items():
            # need to pop out group items so that it wont be reused with fields
            group_field_dict = {
                k: local_field_dict.pop(k)
                for k, v in obj.VIEW.FIELDS.items()
                if v.get("group") == group
            }
            group_ele = E(group_xpath)
            for key, val in group_field_dict.items():
                group_ele.append(E(val.get("xpath")))
            parser_ingest.append(group_ele)
    for i, item in enumerate(local_field_dict.items()):
        # i is the index and item will be taple of field key and value
        field_dict = item[1]
        if "table" in field_dict:
            # handle nested table/view
            child_table = field_dict.get("table")
            parser_ingest.insert(i + 1, generate_sax_parser_input(child_table))
        else:
            xpath = field_dict.get("xpath")
            # xpath can be multi level, for ex traffic-statistics/input-pps
            # going in reverse order, for fields example.
            # split xpath in 2 part, search for first part xpath, if exists, append later
            # else continue, and finally add full xpath to parent
            # min-delay: probe-test-global-results/probe-test-generic-results/probe-test-rtt/probe-summary-results/min-delay
            # max-delay: probe-test-global-results/probe-test-generic-results/probe-test-rtt/probe-summary-results/max-delay
            # avg-delay: probe-test-global-results/probe-test-generic-results/probe-test-rtt/probe-summary-results/avg-delay
            # positive-rtt-jitter: probe-test-global-results/probe-test-generic-results/probe-test-positive-round-trip-jitter/probe-summary-results/avg-delay
            # loss-percentage: probe-test-global-results/probe-test-generic-results/loss-percentage
            # current-min-delay: probe-last-test-results/probe-test-generic-results/probe-test-rtt/probe-summary-results/min-delay
            # current-max-delay: probe-last-test-results/probe-test-generic-results/probe-test-rtt/probe-summary-results/max-delay
            # current-avg-delay: probe-last-test-results/probe-test-generic-results/probe-test-rtt/probe-summary-results/avg-delay
            # current-positive-rtt-jitter: probe-last-test-results/probe-test-generic-results/probe-test-positive-round-trip-jitter/probe-summary-results/avg-delay
            # current-loss-percentage: probe-last-test-results/probe-test-generic-results/loss-percentage
            if "/" in xpath:
                tags = xpath.split("/")
                tags_len = len(tags)
                local_elem_to_add = E(tags[-1])
                for i in range(tags_len, 0, -1):
                    xpath = "/".join(tags[:i])
                    local_obj = parser_ingest
                    elem_exists = local_obj.xpath(xpath)
                    if elem_exists:
                        xpath_exists = elem_exists[0]
                        xpath_exists.insert(1, local_elem_to_add)
                        break
                    if local_elem_to_add.tag != tags[i - 1]:
                        local_elem_to_add = E(tags[i - 1], local_elem_to_add)
                else:
                    parser_ingest.insert(1, local_elem_to_add)
            else:
                parser_ingest.insert(i + 1, E(xpath))
    # cases where item is something like
    # item: task-memory-malloc-usage-report/task-malloc-list/task-malloc
    # created filter from last item task-malloc
    # Now add all the tags if present
    for item_tag in item_tags[::-1]:
        parser_ingest = E(item_tag, parser_ingest)
    logger.debug("Generated filter XML is: %s" % etree.tostring(parser_ingest))

    return parser_ingest



from jnpr.junos.exception import ConnectNotMasterError
from jnpr.junos.exception import RpcError


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "RE_hw_mi": "(Routing Engine hardware multi-instance) A boolean "
        "indicating if this is a multi-chassis system.",
        "serialnumber": "A string containing the serial number of the "
        "device's chassis. If there is no chassis serial "
        "number, the serial number of the backplane or "
        "midplane is returned.",
    }


def get_facts(device):
    """
    Gathers facts from the <get-chassis-inventory/> RPC.
    """
    rsp = device.rpc.get_chassis_inventory(normalize=True)
    if rsp.tag == "error":
        raise RpcError()

    if (
        rsp.tag == "output"
        and rsp.text.find("can only be used on the master routing engine") != -1
    ):
        # An error; due to the fact that this RPC can only be executed on the
        # master Routing Engine
        raise ConnectNotMasterError()

    RE_hw_mi = False
    if rsp.tag == "multi-routing-engine-results":
        RE_hw_mi = True

    serialnumber = (
        rsp.findtext(".//chassis[1]/serial-number")
        or rsp.findtext('.//chassis-module[name="Backplane"]/serial-number')
        or rsp.findtext('.//chassis-module[name="Midplane"]/serial-number')
    )

    return {
        "RE_hw_mi": RE_hw_mi,
        "serialnumber": serialnumber,
    }



from jnpr.junos.jxml import strip_comments_transform
import json
from lxml import etree
from copy import deepcopy

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


class TableJSONEncoder(json.JSONEncoder):
    """
    Used to encode Table/View instances into JSON.  See :meth:`Table.to_json`.
    """

    def default(self, obj):
        from jnpr.junos.factory.view import View
        from jnpr.junos.factory.table import Table
        from jnpr.junos.factory.cmdtable import CMDTable

        if isinstance(obj, View):
            obj = dict(obj.items())
        elif isinstance(obj, Table):
            ret = dict()
            for item in obj:
                if isinstance(item.name, list) and len(item.name) == 0:
                    ret.update(item)
                else:
                    ret[str(item.name)] = item
            return ret
        elif isinstance(obj, CMDTable):
            obj = {str(key): val for key, val in obj.items()}
        else:
            obj = super(TableJSONEncoder, self).default(obj)
        return obj


class TableViewJSONEncoder(json.JSONEncoder):
    """
    Used to encode Table/View instances into JSON.  See :meth:`Table.to_json`.
    """

    def default(self, obj):
        from jnpr.junos.factory.view import View
        from jnpr.junos.factory.table import Table

        if isinstance(obj, View):
            obj = {str(obj.name): dict(obj.items())}
        elif isinstance(obj, Table):
            obj = dict((str(item.name), dict(item.items())) for item in obj)
        else:
            obj = super(TableViewJSONEncoder, self).default(obj)
        return obj


class PyEzJSONEncoder(json.JSONEncoder):
    """
    Used to encode facts and rpc instances into JSON.`.
    """

    def default(self, obj):
        from jnpr.junos.facts.swver import version_info

        if isinstance(obj, version_info):
            obj = obj.v_dict
        elif isinstance(obj, etree._Element):

            def recursive_dict(element):
                return (element.tag, dict(map(recursive_dict, element)) or element.text)

            # JSON does not support comments - strip them
            obj = strip_comments_transform(deepcopy(obj)).getroot()
            _, obj = recursive_dict(obj)
        elif isinstance(obj, MutableMapping):
            obj = {k: v for k, v in obj.items()}
        else:
            obj = super(PyEzJSONEncoder, self).default(obj)
        return obj



# Import any exceptions raised in this module.
from jnpr.junos.exception import RpcError


# In general, a "fact" should be a piece of information that does not change
# over the life of a PyEZ connection. While things like mastership state can
# change, they also result in the PyEZ connection being closed, so it can
# be considered a fact. Things like number of routes in the route table or
# system uptime do not maintain the same value over the life of a PyEZ
# connection and should therefore not be considered facts.

# The name of this file should be based on the RPCs which are invoked by this
# file. This avoids accidentally invoking the same RPC multiple times when we
# could invoke it just once and gather multiple facts from the output.

# An import for each fact file must be present in
# lib/jnpr/junos/facts/__init__.py


# The file must include a provide_facts() function
# The provide_facts() function must return a dictionary. The keys of the
# dictionary are each fact that is handled/returned by this module. The value
# of each item in the dictionary is the documentation string describing the
# particular fact.
def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "foo": "The foo information.",
        "bar": "The bar information.",
    }


# The file must include a get_facts(device) function. The get_facts(device)
# function takes a single mandatory device argument which is the Device object
# on which the fact is discovered.
# The get_facts(device) function must return a dictionary with a key for each
# fact that is handled/returned by this module.
# The get_facts(device) function should raise an appropriate exception if the
# response from the device prevents getting value(s) for ALL of the facts
# provided by the module. The example gets the 'foo' and 'bar' facts.
# An exception should only be raised if the error prevents getting the value
# for the 'foo' fact AND prevents getting the value for the 'bar' fact. If
# unable to determine the value for a given fact, then set the value for that
# fact to None.
def get_facts(device):
    """
    Gathers facts from the <get-foo-bar-information/> RPC.
    """
    # Invoke an RPC on device using device.rpc.rpc_name() methods.
    # Avoid using the cli() or shell() methods.
    # Avoid any RPC which requires more than view privileges.
    # Avoid any RPC which doesn't run on non-master REs.
    # Avoid any RPC which takes more than a couple of seconds to execute.
    # Avoid any RPC which MIGHT take longer than the default PyEZ RPC timeout
    # (currently 30 seconds).
    # If there are different ways of gathering the info depending on
    # Junos version, platform, model, or other things, then try the most common
    # way first, catch exceptions, try the second most common, etc.
    # Note that executing an RPC might itself raise an exception. There is no
    # need to catch that exception. It is handled within the FactCache() class.
    # Always pass the normalize=True argument when invoking an RPC to avoid any
    # potential white-space problems in the RPC response.
    rsp = device.rpc.get_foo_bar_information(normalize=True)

    # Handle any exceptional situations unique to this RPC which prevent
    # getting values for ALL of the facts provided by this module.
    # These are examples which MAY or MAY NOT apply to your particular RPC.
    # Don't just blindly leave these in.
    if rsp.tag == "error":
        raise RpcError()

    # An example of a boolean fact. False if the top-level tag is not 'foo'.
    foo = False
    if rsp.tag == "foo":
        foo = True

    # An example of a string value which might be found at various levels or
    # locations within the response hierarchy.
    bar = (
        rsp.findtext(".//chassis[1]/bar")
        or rsp.findtext('.//chassis-module[name="Backplane"]/bar')
        or rsp.findtext('.//chassis-module[name="Midplane"]/bar')
    )

    return {
        "foo": foo,
        "bar": bar,
    }



# 3rd-party
from lxml.builder import E

# local
from jnpr.junos.cfg import Resource
from jnpr.junos import JXML
from jnpr.junos.cfg.phyport.base import PhyPortBase


class PhyPortSwitch(PhyPortBase):
    PORT_SPEED = {
        "auto": "auto-negotiation",
        "10m": "ethernet-10m",
        "100m": "ethernet-100m",
        "1g": "ethernet-1g",
    }

    # -----------------------------------------------------------------------
    # XML readers
    # -----------------------------------------------------------------------

    def _xml_to_py(self, has_xml, has_py):
        PhyPortBase._xml_to_py(self, has_xml, has_py)

        # speed, duplex, loopback are all under 'ether-options'
        ethopts = has_xml.find("ether-options")
        if ethopts is None:
            return

        if ethopts.find("loopback") is not None:
            has_py["loopback"] = True

        speed = ethopts.find("speed")
        if speed is not None:
            # take the first child element
            has_py["speed"] = speed[0].tag
            PhyPortBase._set_invert(has_py, "speed", self.PORT_SPEED)

        Resource.copyifexists(ethopts, "link-mode", has_py, "duplex")
        if "duplex" in has_py:
            PhyPortBase._set_invert(has_py, "duplex", self.PORT_DUPLEX)

    # -----------------------------------------------------------------------
    # XML writers
    # -----------------------------------------------------------------------

    def _xml_hook_build_change_begin(self, xml):
        if any([this in self.should for this in ["speed", "duplex", "loopback"]]):
            self._ethopts = E("ether-options")
            xml.append(self._ethopts)

    def _xml_change_speed(self, xml):
        speed_tag = self.PORT_SPEED.get(self.speed)
        add_this = E.speed(JXML.DEL) if speed_tag is None else E.speed(E(speed_tag))
        self._ethopts.append(add_this)
        return True

    def _xml_change_duplex(self, xml):
        value = self.PORT_DUPLEX.get(self.duplex)
        Resource.xml_set_or_delete(self._ethopts, "link-mode", value)
        return True

    def _xml_change_loopback(self, xml):
        self._ethopts.append(Resource.xmltag_set_or_del("loopback", self.loopback))
        return True



from copy import deepcopy
import re

from lxml import etree
from lxml.builder import E
from jnpr.junos.factory.table import Table
from jnpr.junos import jxml
from jnpr.junos.utils.config import Config


class CfgTable(Table):
    __isfrozen = False

    # -----------------------------------------------------------------------
    # CONSTRUCTOR
    # -----------------------------------------------------------------------
    def __init__(self, dev=None, xml=None, path=None, mode=None):
        Table.__init__(self, dev, xml, path)  # call parent constructor

        self._init_get()
        self._data_dict = self.DEFINE  # crutch
        self.ITEM_NAME_XPATH = self._data_dict.get("key", "name")
        self.view = self._data_dict.get("view")
        self._options = self._data_dict.get("options")
        self.mode = mode
        if "set" in self._data_dict:
            Config.__init__(self, dev, mode)  # call parent constructor

            self._init_set()
            if self._view:
                self.fields = self._view.FIELDS.copy()
            else:
                raise ValueError(
                    "%s set table view is not defined.\n" % (self.__class__.__name__)
                )
            if "key-field" in self._data_dict:
                key_name = self._data_dict.get("key-field", None)
                if isinstance(key_name, list):
                    self.key_field = key_name
                elif isinstance(key_name, str):
                    self.key_field = [key_name]
                else:
                    raise TypeError(
                        "Key-field %s is of invalid type %s.\n"
                        % (key_name, type(key_name))
                    )
            else:
                raise ValueError("Table should have key-field attribute defined\n")
            self._type = "set"
            self._init_field()
        else:
            self._type = "get"
        self.ITEM_XPATH = self._data_dict[self._type]

        # no new attributes.
        self._freeze()

    # -----------------------------------------------------------------------
    # PROPERTIES
    # -----------------------------------------------------------------------

    @property
    def required_keys(self):
        """
        return a list of the keys required when invoking :get():
        and :get_keys():
        """
        return self._data_dict.get("required_keys")

    @property
    def keys_required(self):
        """True/False - if this Table requires keys"""
        return self.required_keys is not None

    # -----------------------------------------------------------------------
    # PRIVATE METHODS
    # -----------------------------------------------------------------------
    def _init_get(self):
        self._get_xpath = None

        # for debug purposes
        self._get_cmd = None
        self._get_opt = None

    def _init_set(self):
        self._insert_node = None

        # lxml object of configuration xml
        self._config_xml_req = None

        # To check if field value is set.
        self._is_field_set = False

        # for debug purposes
        self._load_rsp = None
        self._commit_rsp = None

    def _buildxml(self, namesonly=False):
        """
        Return an lxml Element starting with <configuration> and comprised
        of the elements specified in the xpath expression.

        For example, and xpath = 'interfaces/interface' would produce:

         <configuration>
            <interfaces>
              <interface/>
            </interfaces>
          </configuration>

        If :namesonly: is True, then the XML will be encoded only to
        retrieve the XML name keys at the 'end' of the XPath expression.

        This return value can then be passed to dev.rpc.get_config()
        to retrieve the specifc data
        """
        xpath = self._data_dict[self._type]
        self._get_xpath = "//configuration/" + xpath
        top = E("configuration")
        dot = top
        for name in xpath.split("/"):
            dot.append(E(name))
            dot = dot[0]

        if namesonly is True:
            dot.attrib["recurse"] = "false"
        return top

    def _build_config_xml(self, top):
        """
        used to encode the field values into the configuration XML
        for set table,each of the field=<value> pairs are defined by user:
        """
        for field_name, opt in self.fields.items():
            dot = top
            # create an XML element with the key/value
            field_value = getattr(self, field_name, None)
            # If field value is not set ignore it
            if field_value is None:
                continue

            if isinstance(field_value, (list, tuple, set)):
                [self._validate_value(field_name, v, opt) for v in field_value]
            else:
                self._validate_value(field_name, field_value, opt)

            field_dict = self.fields[field_name]

            if "group" in field_dict:
                group_xpath = self._view.GROUPS[field_dict["group"]]
                dot = self._encode_xpath(top, group_xpath.split("/"))

            lxpath = field_dict["xpath"].split("/")
            if len(lxpath) > 1:
                dot = self._encode_xpath(top, lxpath[0 : len(lxpath) - 1])

            add_field = self._grindfield(lxpath[-1], field_value)
            for _add in add_field:
                if len(_add.attrib) > 0:
                    for i in dot.iter():
                        if i.tag == _add.tag:
                            i.attrib.update(_add.attrib)
                            break
                    else:
                        dot.append(_add)
                elif field_name in self.key_field:
                    dot.insert(0, _add)
                else:
                    dot.append(_add)

    def _validate_value(self, field_name, value, opt):
        """
        Validate value set for field against the constraints and
        data type check define in yml table/view defination.

        :param field_name: Name of field as mentioned in yaml table/view
        :param value: Value set by user for field_name.
        :param opt: Dictionary of data type and constraint check.
        :return:
        """

        def _get_field_type(ftype):
            ft = {
                "str": str,
                "int": int,
                "float": float,
                "bool": bool,
            }.get(ftype, None)

            if ft is None:
                raise TypeError("Unsupported type %s\n" % (ftype))
            return ft

        def _validate_enum_value(field_name, value, enum_value):
            if isinstance(enum_value, list):
                if value not in enum_value:
                    raise ValueError(
                        "Invalid value %s assigned " "to field %s" % (value, field_name)
                    )
            elif isinstance(enum_value, str):
                if not value == enum_value:
                    raise ValueError(
                        "Invalid value %s assigned " "to field %s" % (value, field_name)
                    )
            else:
                raise TypeError(
                    "Value of enum should " "be either a string or list of strings.\n"
                )

        def _validate_type(field_name, value, opt):
            if isinstance(opt["type"], dict):
                if "enum" in opt["type"]:
                    _validate_enum_value(field_name, value, opt["type"]["enum"])
                else:
                    # More user defined type check can be added in future.
                    # raise execption for now.
                    raise TypeError("Unsupported type %s\n" % (opt["type"]))

            elif isinstance(opt["type"], str):
                field_type = _get_field_type(opt["type"])
                if not isinstance(value, field_type):
                    raise TypeError(
                        "Invalid value %s asigned to field %s,"
                        " value should be of type %s\n"
                        % (value, field_name, field_type)
                    )
            else:
                raise TypeError(
                    "Invalid value %s, should be either of"
                    " type string or dictionary.\n" % (opt["type"])
                )

        def _validate_min_max_value(field_name, value, opt):
            if isinstance(value, (int, float)):
                if value < opt["minValue"] or value > opt["maxValue"]:
                    raise ValueError(
                        "Invalid value %s assigned "
                        "to field %s.\n" % (value, field_name)
                    )
            elif isinstance(value, str):
                if len(value) < opt["minValue"] or len(value) > opt["maxValue"]:
                    raise ValueError(
                        "Invalid value %s assigned "
                        "to field %s.\n" % (value, field_name)
                    )

        if isinstance(value, dict):
            # in case user want to pass operation attr for ex:
            # <unit operation="delete"/>
            pass
        elif isinstance(value, (list, tuple, dict, set)):
            raise ValueError("%s value is invalid %s\n" % (field_name, value))
        else:
            if "type" in opt:
                _validate_type(field_name, value, opt)
            if ("minValue" or "maxValue") in opt:
                _validate_min_max_value(field_name, value, opt)

    def _grindkey(self, key_xpath, key_value):
        """returns list of XML elements for key values"""
        simple = lambda: [E(key_xpath.replace("_", "-"), key_value)]
        composite = lambda: [
            E(xp.replace("_", "-"), xv) for xp, xv in zip(key_xpath, key_value)
        ]
        return simple() if isinstance(key_xpath, str) else composite()

    def _grindxpath(self, key_xpath, key_value):
        """returns xpath elements for key values"""
        simple = lambda: "[{}='{}']".format(key_xpath.replace("_", "-"), key_value)
        composite = lambda: "[{}]".format(
            " and ".join(
                [
                    "{}='{}'".format(xp.replace("_", "-"), xv)
                    for xp, xv in zip(key_xpath, key_value)
                ]
            )
        )
        return simple() if isinstance(key_xpath, str) else composite()

    def _grindfield(self, xpath, value):
        """returns list of xml elements for field name-value pairs"""
        lst = []
        if isinstance(value, (list, tuple, set)):
            for v in value:
                lst.append(E(xpath.replace("_", "-"), str(v)))
        elif isinstance(value, bool):
            if value is True:
                lst.append(E(xpath.replace("_", "-")))
            elif value is False:
                lst.append(E(xpath.replace("_", "-"), {"operation": "delete"}))
        elif isinstance(value, dict):
            lst.append(E(xpath.replace("_", "-"), value))
        else:
            lst.append(E(xpath.replace("_", "-"), str(value)))
        return lst

    def _encode_requiredkeys(self, get_cmd, kvargs):
        """
        used to encode the required_keys values into the XML get-command.
        each of the required_key=<value> pairs are defined in :kvargs:
        """
        rqkeys = self._data_dict["required_keys"]
        for key_name in self.required_keys:
            # create an XML element with the key/value
            key_value = kvargs.get(key_name)
            if key_value is None:
                raise ValueError("Missing required-key: '%s'" % (key_name))
            key_xpath = rqkeys[key_name]
            add_keylist_xml = self._grindkey(key_xpath, key_value)

            # now link this item into the XML command, where key_name
            # designates the XML parent element
            key_name = key_name.replace("_", "-")
            dot = get_cmd.find(".//" + key_name)
            if dot is None:
                raise RuntimeError(
                    "Unable to find parent XML for key: '%s'" % (key_name)
                )
            for _at, _add in enumerate(add_keylist_xml):
                dot.insert(_at, _add)

            # Add required key values to _get_xpath
            xid = re.search(r"\b{}\b".format(key_name), self._get_xpath).start() + len(
                key_name
            )

            self._get_xpath = (
                self._get_xpath[:xid]
                + self._grindxpath(key_xpath, key_value)
                + self._get_xpath[xid:]
            )

    def _encode_namekey(self, get_cmd, dot, namekey_value):
        """
        encodes the specific namekey_value into the get command so that the
        returned XML configuration is the complete hierarchy of data.
        """
        namekey_xpath = self._data_dict.get("key", "name")
        keylist_xml = self._grindkey(namekey_xpath, namekey_value)
        for _add in keylist_xml:
            dot.append(_add)

    def _encode_getfields(self, get_cmd, dot):
        for field_xpath in self._data_dict["get_fields"]:
            dot.append(E(field_xpath))

    def _encode_xpath(self, top, lst):
        """
        Create xml element hierarchy for given field. Return container
        node to which field and its value will appended as child elements.
        """
        dot = top
        for index in range(1, len(lst) + 1):
            xp = "/".join(lst[0:index])
            if not len(top.xpath(xp)):
                dot.append(E(lst[index - 1]))
            dot = dot.find(lst[index - 1])
        return dot

    def _keyspec(self):
        """returns tuple (keyname-xpath, item-xpath)"""
        return (self._data_dict.get("key", "name"), self._data_dict[self._type])

    def _init_field(self):
        """
        Initialize fields of set table to it's default value
        (if mentioned in yml Table/View) else set to None.
        """
        for fname, opt in self.fields.items():
            self.__dict__[fname] = opt["default"] if "default" in opt else None

    def _mandatory_check(self):
        """Mandatory checks for set table/view"""
        for key in self.key_field:
            value = getattr(self, key)
            if value is None:
                raise ValueError("%s key-field value is not set.\n" % (key))

    def _freeze(self):
        """
        Freeze class object so that user cannot add new attributes (fields).
        """
        self.__isfrozen = True

    def _unfreeze(self):
        """
        Unfreeze class object, should be called from within class only.
        """
        self.__isfrozen = False

    # ----------------------------------------------------------------------
    # reset - Assign 'set' Table field values to default or None
    # ----------------------------------------------------------------------
    def reset(self):
        """
        Initialize fields of set table to it's default value
        (if mentioned in Table/View) else set to None.
        """
        return self._init_field()

    # ----------------------------------------------------------------------
    # get_table_xml - retrieve lxml configuration object for set table
    # ----------------------------------------------------------------------
    def get_table_xml(self):
        """
        It returns lxml object of configuration xml that is generated
        from table data (field=value) pairs. To get a valid xml this
        method should be used after append() is called.
        """
        return self._config_xml_req

    # ----------------------------------------------------------------------
    # append - append Table data to lxml configuration object
    # ----------------------------------------------------------------------
    def append(self):
        """
        It creates lxml nodes with field name as xml tag and its value
        given by user as text of xml node. The generated xml nodes are
        appended to configuration xml at appropriate hierarchy.

        .. warning::
            xml node that are appended cannot be changed later hence
            care should be taken to assign correct value to table fields
            before calling append.
        """

        # mandatory check for 'set' table fields
        self._mandatory_check()

        set_cmd = self._buildxml()
        top = set_cmd.find(self._data_dict[self._type])
        self._build_config_xml(top)
        if self._config_xml_req is None:
            self._config_xml_req = set_cmd
            self._insert_node = top.getparent()
        else:
            self._insert_node.extend(top.getparent())

        self.reset()  # Reset field values
        self._is_field_set = False

    # ----------------------------------------------------------------------
    # get - retrieve Table data
    # ----------------------------------------------------------------------

    def get(self, *vargs, **kvargs):
        """
        Retrieve configuration data for this table.  By default all child
        keys of the table are loaded.  This behavior can be overridden by
        with kvargs['nameonly']=True

        :param str vargs[0]: identifies a unique item in the table,
          same as calling with :kvargs['key']: value

        :param str namesonly:
          *OPTIONAL* True/False*, when set to True will cause only the
          the name-keys to be retrieved.

        :param str key:
          *OPTIONAL* identifies a unique item in the table

        :param dict options:
          *OPTIONAL* options to pass to get-configuration.  By default
          {'inherit': 'inherit', 'groups': 'groups'} is sent.
        """
        if self._lxml is not None:
            return self

        if self._path is not None:
            # for loading from local file-path
            self.xml = etree.parse(self._path).getroot()
            return self

        if self.keys_required is True and not len(kvargs):
            raise ValueError("This table has required-keys\n", self.required_keys)

        self._clearkeys()

        # determine if we need to get only the names of keys, or all of the
        # hierarchical data from the config.  The caller can explicitly set
        # :namesonly: in the call.

        if "namesonly" in kvargs:
            namesonly = kvargs.get("namesonly")
        else:
            namesonly = False

        get_cmd = self._buildxml(namesonly=namesonly)

        # if this table requires additional keys, for the hierarchical
        # use-cases then make sure these are provided by the caller. Then
        # encode them into the 'get-cmd' XML

        if self.keys_required is True:
            self._encode_requiredkeys(get_cmd, kvargs)

        try:
            # see if the caller provided a named item.  this must
            # be an actual name of a thing, and not an index number.
            # ... at least for now ...
            named_item = kvargs.get("key") or vargs[0]
            dot = get_cmd.find(self._data_dict[self._type])
            self._encode_namekey(get_cmd, dot, named_item)

            if "get_fields" in self._data_dict:
                self._encode_getfields(get_cmd, dot)

        except:
            # caller not requesting a specific table item
            pass

        # Check for options in get
        if "options" in kvargs:
            options = kvargs.get("options") or {}
        else:
            if self._options is not None:
                options = self._options
            else:
                options = jxml.INHERIT_GROUPS

        # for debug purposes
        self._get_cmd = get_cmd
        self._get_opt = options

        # retrieve the XML configuration
        # Check to see if running on box
        if self._dev.ON_JUNOS:
            try:
                from junos import Junos_Configuration

                # If part of commit script use the context
                if Junos_Configuration is not None:
                    # Convert the onbox XML to ncclient reply
                    config = jxml.conf_transform(
                        deepcopy(jxml.cscript_conf(Junos_Configuration)),
                        subSelectionXPath=self._get_xpath,
                    )
                    self.xml = config.getroot()
                else:
                    self.xml = self.RPC.get_config(get_cmd, options=options)
            # If onbox import missing fallback to RPC - possibly raise
            # exception in future
            except ImportError:
                self.xml = self.RPC.get_config(get_cmd, options=options)
        else:
            self.xml = self.RPC.get_config(get_cmd, options=options)

        # return self for call-chaining, yo!
        return self

    # -----------------------------------------------------------------------
    # set - configure Table data in running configuration.
    # -----------------------------------------------------------------------

    def set(self, **kvargs):
        """
        Load configuration data in running db.
        It performs following operation in sequence.

        * lock(): Locks candidate configuration db.
        * load(): Load structured configuration xml in candidate db.
        * commit(): Commit configuration to runnning db.
        * unlock(): Unlock candidate db.

        This method should be used after append() is called to
        get the desired results.

        :param bool overwrite:
          Determines if the contents completely replace the existing
          configuration.  Default is ``False``.

        :param bool merge:
          If set to ``True`` will set the load-config action to merge.
          the default load-config action is 'replace'

        :param str comment: If provided logs this comment with the commit.

        :param int confirm: If provided activates confirm safeguard with
                            provided value as timeout (minutes).

        :param int timeout: If provided the command will wait for completion
                            using the provided value as timeout (seconds).
                            By default the device timeout is used.

        :param bool sync: On dual control plane systems, requests that
                            the candidate configuration on one control plane
                            be copied to the other control plane, checked for
                            correct syntax, and committed on both Routing
                            Engines.

        :param bool force_sync: On dual control plane systems, forces the
                            candidate configuration on one control plane
                            to be copied to the other control plane.

        :param bool full: When true requires all the daemons to check and
                          evaluate the new configuration.

        :param bool detail: When true return commit detail as XML

        :returns: Class object:

        :raises: ConfigLoadError:
                    When errors detected while loading
                    configuration. You can use the Exception errs
                    variable to identify the specific problems

                CommitError:
                    When errors detected in candidate configuration.
                    You can use the Exception errs variable
                    to identify the specific problems

                RuntimeError:
                    If field value is set and append() is not
                    invoked before calling this method, it will
                    raise an exception with appropriate error
                    message.

        .. warning::
            If the function does not receive a reply prior to the timeout
            a RpcTimeoutError will be raised.  It is possible the commit
            was successful.  Manual verification may be required.
        """
        if self._is_field_set:
            raise RuntimeError(
                "Field value is changed, append() " "must be called before set()"
            )

        self.lock()

        try:
            # Invoke config class load() api, with xml object.
            self._load_rsp = super(CfgTable, self).load(self._config_xml_req, **kvargs)
            self._commit_rsp = self.commit(**kvargs)
        finally:
            self.unlock()

        return self

    # -----------------------------------------------------------------------
    # OVERLOADS
    # -----------------------------------------------------------------------

    def __setitem__(self, t_field, value):
        """
        implements []= to set Field value
        """
        if t_field in self.fields:
            # pass 'up' to standard setattr method
            self.__setattr__(t_field, value)
        else:
            raise ValueError("Unknown field: %s" % (t_field))

    def __setattr__(self, attribute, value):
        if self.__isfrozen and not hasattr(self, attribute):
            raise ValueError("Unknown field: %s" % (attribute))
        else:
            # pass 'up' to standard setattr method
            object.__setattr__(self, attribute, value)
            if hasattr(self, "fields") and attribute in self.fields:
                object.__setattr__(self, "_is_field_set", True)

    def __enter__(self):
        return super(CfgTable, self).__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return super(CfgTable, self).__exit__(exc_type, exc_val, exc_tb)

    # -----------------------------------------------------------------------
    # load - configure Table data in candidate configuration.
    # -----------------------------------------------------------------------

    def load(self, **kvargs):
        """
        Load configuration xml having table data (field=value)
        in candidate db.
        This method should be used after append() is called to
        get the desired results.

        :param bool overwrite:
          Determines if the contents completely replace the existing
          configuration.  Default is ``False``.

        :param bool merge:
          If set to ``True`` will set the load-config action to merge.
          the default load-config action is 'replace'

        :returns: Class object.

        :raises: ConfigLoadError:
                    When errors detected while loading
                    configuration. You can use the Exception errs
                    variable to identify the specific problems
                 RuntimeError:
                    If field value is set and append() is not
                    invoked before calling this method, it will
                    raise an exception with appropriate error
                    message.
        """
        if self._is_field_set:
            raise RuntimeError(
                "Field value is changed, append() " "must be called before load()"
            )

        # pass up to config class load() api, with xml object as vargs[0].
        self._load_rsp = super(CfgTable, self).load(self._config_xml_req, **kvargs)
        return self



import re
from jnpr.junos import jxml
from jnpr.junos import jxml as JXML
from lxml.etree import _Element
from ncclient.operations.rpc import RPCError


class FactLoopError(RuntimeError):
    """
    Generated when there is a loop in fact gathering.
    """

    pass


class RpcError(Exception):
    """
    Parent class for all junos-pyez RPC Exceptions
    """

    def __init__(self, cmd=None, rsp=None, errs=None, dev=None, timeout=None, re=None):
        """
        :cmd: is the rpc command
        :rsp: is the rpc response (after <rpc-reply>)
        :errs: is a list of dictionaries of extracted <rpc-error> elements.
        :dev: is the device rpc was executed on
        :timeout: is the timeout value of the device
        :re: is the RE or member exception occured on
        """
        self.cmd = cmd
        self.rsp = rsp
        self.dev = dev
        self.timeout = timeout
        self.re = re
        self.rpc_error = None
        self.xml = rsp
        # To handle errors coming from ncclient, Here errs is list of RPCError
        if isinstance(errs, RPCError) and hasattr(errs, "errors"):
            self.errs = [JXML.rpc_error(error.xml) for error in errs.errors]
            for error in errs.errors:
                if error.severity == "error":
                    self.rsp = JXML.remove_namespaces(error.xml)
                    break
            else:
                if errs.severity == "warning":
                    for error in errs.errors:
                        if error.severity == "warning":
                            self.rsp = JXML.remove_namespaces(error.xml)
                            break
            self.message = errs.message
        else:
            self.errs = errs
            self.message = (
                "\n".join(
                    [
                        "%s: %s" % (err["severity"].strip(), err["message"].strip())
                        for err in errs
                        if err["message"] is not None and err["severity"] is not None
                    ]
                )
                if isinstance(errs, list)
                else ""
            )

        if isinstance(self.rsp, _Element):
            self.rpc_error = jxml.rpc_error(self.rsp)
            self.message = self.message or self.rpc_error["message"]
            if self.errs is None or not isinstance(self.errs, list):
                self.errs = [self.rpc_error]

    def __repr__(self):
        """
        pprints the response XML attribute
        """
        if self.rpc_error is not None:
            return "{}(severity: {}, bad_element: {}, message: {})".format(
                self.__class__.__name__,
                self.rpc_error["severity"],
                self.rpc_error["bad_element"],
                self.message,
            )
        else:
            return self.__class__.__name__

    __str__ = __repr__


class CommitError(RpcError):
    """
    Generated in response to a commit-check or a commit action.
    """

    def __init__(self, rsp, cmd=None, errs=None):
        RpcError.__init__(self, cmd, rsp, errs)

    def __repr__(self):
        return "{}(edit_path: {}, bad_element: {}, message: {})".format(
            self.__class__.__name__,
            self.rpc_error["edit_path"],
            self.rpc_error["bad_element"],
            self.message,
        )

    __str__ = __repr__


class ConfigLoadError(RpcError):
    """
    Generated in response to a failure when loading a configuration.
    """

    def __init__(self, rsp, cmd=None, errs=None):
        RpcError.__init__(self, cmd, rsp, errs)

    def __repr__(self):
        return "{}(severity: {}, bad_element: {}, message: {})".format(
            self.__class__.__name__,
            self.rpc_error["severity"],
            self.rpc_error["bad_element"],
            self.message,
        )

    __str__ = __repr__


class LockError(RpcError):
    """
    Generated in response to attempting to take an exclusive
    lock on the configuration database.
    """

    def __init__(self, rsp):
        RpcError.__init__(self, rsp=rsp)


class UnlockError(RpcError):
    """
    Generated in response to attempting to unlock the
    configuration database.
    """

    def __init__(self, rsp):
        RpcError.__init__(self, rsp=rsp)


class PermissionError(RpcError):
    """
    Generated in response to invoking an RPC for which the
    auth user does not have user-class permissions.

    PermissionError.message gives you the specific RPC that cause
    the exceptions
    """

    def __init__(self, rsp, cmd=None, errs=None):
        RpcError.__init__(self, cmd=cmd, rsp=rsp, errs=errs)
        self.message = rsp.findtext(".//bad-element")


class RpcTimeoutError(RpcError):
    """
    Generated in response to a RPC execution timeout.
    """

    def __init__(self, dev, cmd, timeout):
        RpcError.__init__(self, dev=dev, cmd=cmd, timeout=timeout)

    def __repr__(self):
        return "{}(host: {}, cmd: {}, timeout: {})".format(
            self.__class__.__name__, self.dev.hostname, self.cmd, self.timeout
        )

    __str__ = __repr__


class SwRollbackError(RpcError):
    """
    Generated in response to a SW rollback error.
    """

    def __init__(self, rsp, re=None):
        RpcError.__init__(self, re=re, rsp=rsp)

    def __repr__(self):
        if self.re:
            return "{}(re: {}, output: {})".format(
                self.__class__.__name__, self.re, self.rsp
            )
        else:
            return "{}(output: {})".format(self.__class__.__name__, self.rsp)

    __str__ = __repr__


# ================================================================
# ================================================================
#                    Connection Exceptions
# ================================================================
# ================================================================


class ConnectError(Exception):
    """
    Parent class for all connection related exceptions
    """

    def __init__(self, dev, msg=None):
        self.dev = dev
        self._orig = msg

    @property
    def user(self):
        """login user-name"""
        return self.dev.user

    @property
    def host(self):
        """login host name/ipaddr"""
        return self.dev.hostname

    @property
    def port(self):
        """login SSH port"""
        return self.dev._port

    @property
    def msg(self):
        """login SSH port"""
        return self._orig

    def __repr__(self):
        if self._orig:
            return "{}(host: {}, msg: {})".format(
                self.__class__.__name__, self.dev.hostname, self._orig
            )
        else:
            return "{}({})".format(self.__class__.__name__, self.dev.hostname)

    __str__ = __repr__


class ProbeError(ConnectError):
    """
    Generated if auto_probe is enabled and the probe action fails
    """

    pass


class ConnectAuthError(ConnectError):
    """
    Generated if the user-name, password is invalid
    """

    pass


class ConnectTimeoutError(ConnectError):
    """
    Generated if the NETCONF session fails to connect, could
    be due to the fact the device is not ip reachable; bad
    ipaddr or just due to routing
    """

    pass


class ConnectUnknownHostError(ConnectError):
    """
    Generated if the specific hostname does not DNS resolve
    """

    pass


class ConnectRefusedError(ConnectError):
    """
    Generated if the specified host denies the NETCONF; could
    be that the services is not enabled, or the host has
    too many connections already.
    """

    pass


class ConnectNotMasterError(ConnectError):
    """
    Generated if the connection is made to a non-master
    routing-engine.  This could be a backup RE on an MX
    device, or a virtual-chassis member (linecard), for example
    """

    pass


class ConnectClosedError(ConnectError):
    """
    Generated if connection unexpectedly closed
    """

    def __init__(self, dev):
        ConnectError.__init__(self, dev=dev)
        dev.connected = False


class JSONLoadError(Exception):
    """
    Generated if json content of rpc reply fails to load
    """

    def __init__(self, exception, rpc_content):
        self.ex_msg = str(exception)
        self.rpc_content = rpc_content
        self.offending_line = ""
        obj = re.search(r"line (\d+)", self.ex_msg)
        if obj:
            line_no = int(obj.group(1))
            rpc_lines = rpc_content.splitlines()
            for line in range(line_no - 3, line_no + 2):
                self.offending_line += "%s: %s\n" % (line + 1, rpc_lines[line])

    def __repr__(self):
        if self.offending_line:
            return (
                "{}(reason: {}, \nThe offending config appears "
                "to be: \n{})".format(
                    self.__class__.__name__, self.ex_msg, self.offending_line
                )
            )
        else:
            return "{}(reason: {})".format(self.__class__.__name__, self.ex_msg)

    __str__ = __repr__



import serial
import re
import six
from time import sleep
from datetime import datetime, timedelta

from jnpr.junos.transport.tty import Terminal

# -------------------------------------------------------------------------
# Terminal connection over SERIAL CONSOLE
# -------------------------------------------------------------------------

_PROMPT = re.compile(six.b("|").join([six.b(i) for i in Terminal._RE_PAT]))


class Serial(Terminal):
    def __init__(self, port="/dev/ttyUSB0", **kvargs):
        """
        :port:
          the serial port, defaults to USB0 since this

        :kvargs['timeout']:
          this is the tty read polling timeout.
          generally you should not have to tweak this.
        """
        # initialize the underlying TTY device

        self.port = port
        self._ser = serial.Serial()
        self._rx = self._ser
        self._ser.port = port
        self._ser.timeout = kvargs.get("timeout", self.TIMEOUT)

        self._tty_name = self.port

        Terminal.__init__(self, **kvargs)

    # -------------------------------------------------------------------------
    # I/O open close called from Terminal class
    # -------------------------------------------------------------------------

    def _tty_open(self):
        try:
            self._ser.open()
        except OSError as err:
            raise RuntimeError("open_failed:{}".format(err.strerror))
        self.write("\n\n\n")  # hit <ENTER> a few times, yo!

    def _tty_close(self):
        self._ser.flush()
        self._ser.close()

    # -------------------------------------------------------------------------
    # I/O read and write called from Terminal class
    # -------------------------------------------------------------------------

    def write(self, content):
        """write content + <RETURN>"""
        self._ser.write(six.b(content + "\n"))
        self._ser.flush()

    def rawwrite(self, content):
        self._ser.write(content)

    def read(self):
        """read a single line"""
        return self._ser.readline()

    def read_prompt(self):
        """
        reads text from the serial console (using readline) until
        a match is found against the :expect: regular-expression object.
        When a match is found, return a tuple(<text>,<found>) where
        <text> is the complete text and <found> is the name of the
        regular-expression group. If a timeout occurs, then return
        the tuple(None,None).
        """
        rxb = six.b("")
        mark_start = datetime.now()
        mark_end = mark_start + timedelta(seconds=self.EXPECT_TIMEOUT)

        while datetime.now() < mark_end:
            sleep(0.1)  # do not remove
            line = self._ser.readline()
            if not line:
                continue
            rxb += line
            found = _PROMPT.search(rxb)
            if found is not None:
                break  # done reading
        else:
            # exceeded the while loop timeout
            return (None, None)

        return (rxb, found.lastgroup)



from jnpr.junos.ofacts.chassis import facts_chassis
from jnpr.junos.ofacts.routing_engines import facts_routing_engines
from jnpr.junos.ofacts.personality import facts_personality
from jnpr.junos.ofacts.swver import facts_software_version
from jnpr.junos.ofacts.ifd_style import facts_ifd_style
from jnpr.junos.ofacts.switch_style import facts_switch_style
from jnpr.junos.ofacts.session import facts_session
from jnpr.junos.ofacts.srx_cluster import facts_srx_cluster
from jnpr.junos.ofacts.domain import facts_domain

FACT_LIST = [
    facts_chassis,  # first
    facts_routing_engines,  # second
    facts_personality,  # third
    facts_srx_cluster,  # four
    facts_software_version,  # fifth
    facts_domain,
    facts_ifd_style,
    facts_switch_style,
    facts_session,
]

__all__ = ["FACT_LIST"]



import re
from jnpr.junos.facts.swver import version_info


def _get_swver(dev, facts):
    # See if we're VC Capable
    if facts["vc_capable"] is True:
        try:
            return dev.rpc.cli("show version all-members", format="xml")
        except:
            pass
    try:
        return dev.rpc.cli("show version invoke-on all-routing-engines", format="xml")
    except:
        return dev.rpc.get_software_information()


def facts_software_version(junos, facts):
    """
    The following facts are required:
        facts['master']

    The following facts are assigned:
        facts['hostname']
        facts['version']
        facts['version_<RE#>'] for each RE in dual-RE, cluster or VC system
        facts['version_info'] for master RE
    """

    x_swver = _get_swver(junos, facts)

    if not facts.get("model"):
        # try to extract the model from the version information
        facts["model"] = x_swver.findtext(".//product-model")

    # ------------------------------------------------------------------------
    # extract the version information out of the RPC response
    # ------------------------------------------------------------------------

    f_master = facts.get("master", "RE0")

    if x_swver.tag == "multi-routing-engine-results":
        # we need to find/identify each of the routing-engine (CPU) versions.
        if len(x_swver.xpath("./multi-routing-engine-item")) > 1:
            facts["2RE"] = True
        versions = []

        if isinstance(f_master, list):
            xpath = (
                './multi-routing-engine-item[re-name="{}"]/software-'
                "information/host-name".format(f_master[0].lower())
            )
        else:
            xpath = (
                './multi-routing-engine-item[re-name="{}"'
                "]/software-information/host-name".format(f_master.lower())
            )

        facts["hostname"] = x_swver.findtext(xpath)
        if facts["hostname"] is None:
            # then there the re-name is not what we are expecting; we should
            # handle this better, eh?  For now, just assume there is one
            # software-information element and take that host-name. @@@ hack.
            facts["hostname"] = x_swver.findtext(".//software-information/host-name")

        for re_sw in x_swver.xpath(".//software-information"):
            re_name = re_sw.xpath("preceding-sibling::re-name")[0].text

            # handle the cases where the "RE name" could be things like
            # "FPC<n>" or "ndoe<n>", and normalize to "RE<n>".
            re_name = re.sub(r"(\w+)(\d+)", "RE\\2", re_name)

            # First try the <junos-version> tag present in >= 15.1
            swinfo = re_sw.findtext("junos-version", default=None)
            if not swinfo:
                # For < 15.1, get version from the "junos" package.
                pkginfo = re_sw.xpath(
                    "package-information[normalize-space(name)=" '"junos"]/comment'
                )[0].text
                try:
                    swinfo = re.findall(r"\[(.*)\]", pkginfo)[0]
                except:
                    swinfo = "0.0I0.0"
            versions.append((re_name.upper(), swinfo))

        # now add the versions to the facts <dict>
        for re_ver in versions:
            facts["version_" + re_ver[0]] = re_ver[1]

        if f_master is not None:
            master = f_master[0] if isinstance(f_master, list) else f_master
            if "version_" + master in facts:
                facts["version"] = facts["version_" + master]
            else:
                facts["version"] = versions[0][1]
        else:
            facts["version"] = versions[0][1]

    else:
        # single-RE
        facts["hostname"] = x_swver.findtext("host-name")

        # First try the <junos-version> tag present in >= 15.1
        swinfo = x_swver.findtext(".//junos-version", default=None)
        if not swinfo:
            # For < 15.1, get version from the "junos" package.
            pkginfo = x_swver.xpath(
                './/package-information[normalize-space(name)="junos"]/comment'
            )[0].text
            try:
                swinfo = re.findall(r"\[(.*)\]", pkginfo)[0]
            except:
                swinfo = "0.0I0.0"
        facts["version"] = swinfo

    # ------------------------------------------------------------------------
    # create a 'version_info' object based on the master version
    # ------------------------------------------------------------------------

    facts["version_info"] = version_info(facts["version"])



import os
import sys
from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_loader

import yaml
import yamlordereddictloader
from jnpr.junos.factory.factory_loader import FactoryLoader

__all__ = []


class MetaPathFinder(MetaPathFinder):
    def find_spec(self, fullname, path=None, target=None):
        mod = fullname.split(".")[-1]
        if mod in [
            os.path.splitext(i)[0] for i in os.listdir(os.path.dirname(__file__))
        ]:
            return spec_from_loader(fullname, MetaPathLoader(fullname))


class MetaPathLoader(Loader):
    def __init__(self, fullname):
        self.fullname = fullname
        self.modules = {}

    def exec_module(self, module):
        if self.fullname in self.modules:
            return self.modules[self.fullname]

        mod = self.fullname.split(".")[-1]
        with open(os.path.join(os.path.dirname(__file__), mod + ".yml"), "r") as stream:
            try:
                modules = FactoryLoader().load(
                    yaml.load(stream, Loader=yamlordereddictloader.Loader)
                )
            except yaml.YAMLError as exc:
                raise ImportError("%s is not loaded" % mod)

        for k, v in modules.items():
            setattr(module, k, v)

        self.modules[self.fullname] = module

        return module


sys.meta_path.insert(0, MetaPathFinder())



import re


class version_info(object):
    def __init__(self, verstr):
        """verstr - version string"""
        m1 = re.match("(.*?)([RBIXSF-])(.*)", verstr)
        self.type = m1.group(2)

        self.major = tuple(map(int, m1.group(1).split(".")))  # creates tuyple
        after_type = m1.group(3).split(".")
        self.minor = after_type[0]

        if "X" == self.type:
            # assumes form similar to "45-D10", so extract the bits from this
            xm = re.match(r"(\d+)-(\w)(\d+)", self.minor)
            if xm is not None:
                self.minor = tuple([int(xm.group(1)), xm.group(2), int(xm.group(3))])
                if len(after_type) < 2:
                    self.build = None
                else:
                    try:
                        # handling case for EVO format X100-202310100600.0-EVO
                        self.build = int(after_type[1])
                    except:
                        self.build = None

            # X type not hyphen format, perhaps "11.4X12.1", just extract
            # build rev or set None
            else:
                if len(after_type) < 2:
                    self.build = None
                else:
                    try:
                        # handling case for EVO format X50.17-EVO
                        self.build = int(after_type[1])
                    except:
                        self.build = None

        elif ("I" == self.type) or ("-" == self.type):
            self.type = "I"
            try:
                # assumes that we have a build/spin, but not numeric
                self.build = after_type[1]
            except:
                self.build = None
        else:
            try:
                self.build = int(after_type[1])  # assumes numeric build/spin
            except:
                self.build = after_type[0]  # non-numeric

        self.as_tuple = self.major + tuple([self.type, self.minor, self.build])
        self.v_dict = {
            "major": self.major,
            "type": self.type,
            "minor": self.minor,
            "build": self.build,
        }

    def __iter__(self):
        for key in self.v_dict:
            yield key, self.v_dict[key]

    def __repr__(self):
        retstr = (
            "junos.version_info(major={major}, type={type},"
            " minor={minor}, build={build})".format(
                major=self.major, type=self.type, minor=self.minor, build=self.build
            )
        )
        return retstr

    def _cmp_tuple(self, other):
        length = len(self) if len(self) < len(other) else len(other)
        return self.as_tuple[0:length]

    def __len__(self):
        length = 0
        for component in self.as_tuple:
            if component is None:
                return length
            else:
                length += 1
        return length

    def __lt__(self, other):
        return self._cmp_tuple(other) < other

    def __le__(self, other):
        return self._cmp_tuple(other) <= other

    def __gt__(self, other):
        return self._cmp_tuple(other) > other

    def __ge__(self, other):
        return self._cmp_tuple(other) >= other

    def __eq__(self, other):
        return self._cmp_tuple(other) == other

    def __ne__(self, other):
        return self._cmp_tuple(other) != other


def version_yaml_representer(dumper, version):
    return dumper.represent_mapping("tag:yaml.org,2002:map", version.v_dict)


def provides_facts():
    """
    Doesn't really provide any facts.
    """
    return {}


def get_facts(device):
    """
    Doesn't get any facts.
    """
    return {}



# stdlib
from functools import wraps
import re
import sys

from lxml import etree
from ncclient.operations.rpc import RPCError
from ncclient.xml_ import NCElement
from jnpr.junos import jxml as JXML


def timeoutDecorator(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        if "dev_timeout" in kwargs:
            try:
                dev = args[0].dev
            except:
                dev = args[0]
            restore_timeout = dev.timeout
            dev.timeout = kwargs.pop("dev_timeout", None)
            try:
                result = function(*args, **kwargs)
                dev.timeout = restore_timeout
                return result
            except Exception:
                dev.timeout = restore_timeout
                raise
        else:
            try:
                return function(*args, **kwargs)
            except Exception:
                raise

    return wrapper


def normalizeDecorator(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        if "normalize" in kwargs:
            normalize = kwargs.pop("normalize", None)
            try:
                dev = args[0].dev
            except:
                dev = args[0]

            if dev._normalize != normalize:
                restore_transform = dev.transform

                if normalize is False:
                    try:
                        dev.transform = dev._nc_transform
                        result = function(*args, **kwargs)
                        dev.transform = restore_transform
                        return result
                    except Exception:
                        dev.transform = restore_transform
                        raise
                else:
                    try:
                        dev.transform = dev._norm_transform
                        result = function(*args, **kwargs)
                        dev.transform = restore_transform
                        return result
                    except Exception:
                        dev.transform = restore_transform
                        raise
            else:
                try:
                    return function(*args, **kwargs)
                except Exception:
                    raise
        else:
            try:
                return function(*args, **kwargs)
            except Exception:
                raise

    return wrapper


def ignoreWarnDecorator(function):
    """
    Ignore warnings if all <rpc-error> elements are at severity 'warning' and
    match one of the values of the ignore_warning argument.

    For example::
        dev.rpc.get(ignore_warning=True)
        dev.rpc.get(ignore_warning='vrrp subsystem not running')
        dev.rpc.get(ignore_warning=['vrrp subsystem not running',
                                    'statement not found'])
        cu.load(cnf, ignore_warning='statement not found')

    :ignore_warning: A boolean, string or list of string.
        If the value is True, it will ignore all warnings regarldess of the
        warning message. If the value is a string, it will ignore warning(s) if
        the message of each warning matches the string. If the value is a list
        of strings, ignore warning(s) if the message of each warning matches at
        least one of the strings in the list.

    .. note::
            When the value of ignore_warning is a string, or list of strings,
            the string is actually used as a case-insensitive regular
            expression pattern. If the string contains only alpha-numeric
            characters, as shown in the above examples, this results in a
            case-insensitive substring match. However, any regular expression
            pattern supported by the re library may be used for more
            complicated match conditions.
    """

    @wraps(function)
    def wrapper(self, *args, **kwargs):
        ignore_warning = kwargs.pop("ignore_warning", False)
        rsp = None
        try:
            rsp = function(self, *args, **kwargs)
        except RPCError as ex:
            if hasattr(ex, "xml") and ignore_warning:
                if hasattr(ex, "errors"):
                    errors = ex.errors
                else:
                    errors = [ex]
                for err in errors:
                    if err.severity == "warning":
                        if (
                            sys.version < "3"
                            and isinstance(ignore_warning, (str, unicode))
                        ) or (sys.version >= "3" and isinstance(ignore_warning, str)):
                            if not re.search(ignore_warning, err.message, re.I):
                                # Message did not match.
                                raise ex
                        elif isinstance(ignore_warning, list):
                            for warn_msg in ignore_warning:
                                if re.search(warn_msg, err.message, re.I):
                                    # Warning matches.
                                    # Break skips else.
                                    break
                            else:
                                # Message didn't match any of the
                                # ignore_warn pattern values.
                                raise ex
                    else:
                        # Not a warning (probably an error).
                        raise ex
                # Every err was a warning that matched ignore_warning.
                # Prepare the response which will get returned.
                # ex.xml contains the raw xml response which was
                # received, but might be rooted at an <rpc-error> element.
                # Set rsp to the root <rpc-reply> element.
                rsp = ex.xml.getroottree().getroot()
                # 1) A normal response has been run through the XSLT
                #    transformation, but ex.xml has not. Do that now.
                encode = None if sys.version < "3" else "unicode"
                rsp = NCElement(
                    etree.tostring(rsp, encoding=encode), self.transform()
                )._NCElement__doc
                # 2) Now remove all of the <rpc-error> elements from
                #    the response. We've already confirmed they are
                #    all warnings
                rsp = etree.fromstring(str(JXML.strip_rpc_error_transform(rsp)))
            else:
                # ignore_warning was false, or an RPCError which doesn't have
                #  an XML attribute. Raise it up for the caller to deal with.
                raise ex
        return rsp

    return wrapper


def checkSAXParserDecorator(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        # args[0] is self
        use_filter = kwargs.pop("use_filter", args[0]._use_filter)
        restore_value = args[0]._use_filter
        args[0]._use_filter = use_filter
        try:
            if args[0].D != None:
                func = args[0].D.transform
            result = function(*args, **kwargs)
            args[0]._use_filter = restore_value
            if args[0].D != None:
                args[0].D.transform = func
            return result
        except Exception:
            args[0]._use_filter = restore_value
            raise

    return wrapper



import re as RE


def _get_vc_status(dev, facts):
    try:
        rsp = dev.rpc.get_virtual_chassis_information()
        # MX issue where command returns, but without content
        if rsp is not True:
            facts["vc_capable"] = True
            return rsp
        else:
            facts["vc_capable"] = False
            return None
    except:
        facts["vc_capable"] = False
        return None


def facts_routing_engines(junos, facts):
    re_facts = ["mastership-state", "status", "model", "up-time", "last-reboot-reason"]

    master = []

    vc_info = _get_vc_status(junos, facts)

    if vc_info is not None:
        facts["vc_mode"] = vc_info.findtext(".//virtual-chassis-mode")
        if (
            len(vc_info.xpath(".//virtual-chassis-id-information" "[@style='fabric']"))
            > 0
        ):
            facts["vc_fabric"] = True
        vc_list = vc_info.xpath(
            ".//member-role[starts-with(.,'Master') " "or starts-with(.,'Backup')]"
        )
        if len(vc_list) > 1:
            facts["2RE"] = True
        for member_id in vc_info.xpath(
            ".//member-role[starts-with(.,'Master')]" "/preceding-sibling::member-id"
        ):
            master.append("RE{}".format(member_id.text))

    try:
        re_info = junos.rpc.get_route_engine_information()
    except:
        # this means that the RPC failed.  this should "never"
        # happen, but we will trap it cleanly for now
        return

    re_list = re_info.xpath(".//route-engine")

    if len(re_list) > 1:
        facts["2RE"] = True

    for re in re_list:
        x_re_name = re.xpath("ancestor::multi-routing-engine-item/re-name")

        if not x_re_name:
            # not a multi-instance routing engine platform, but could
            # have multiple RE slots
            re_name = "RE"
            x_slot = re.find("slot")
            slot_id = x_slot.text if x_slot is not None else "0"
            re_name = re_name + slot_id
        else:
            # multi-instance routing platform
            m = RE.search(r"(\d)", x_re_name[0].text)
            if vc_info is not None:
                # => RE0-RE0 | RE0-RE1
                re_name = "RE{}-RE{}".format(m.group(0), re.find("slot").text)
            else:
                re_name = "RE" + m.group(0)  # => RE0 | RE1

        re_fd = {}
        facts[re_name] = re_fd
        for factoid in re_facts:
            x_f = re.find(factoid)
            if x_f is not None:
                re_fd[factoid.replace("-", "_")] = x_f.text

        if vc_info is None and "mastership_state" in re_fd:
            if facts[re_name]["mastership_state"] == "master":
                master.append(re_name)

    # --[ end for-each 're' ]-------------------------------------------------

    len_master = len(master)
    if len_master > 1:
        facts["master"] = master
    elif len_master == 1:
        facts["master"] = master[0]



import re
import sys
from lxml import etree
from lxml.builder import E
from jnpr.junos import jxml as JXML


class _RpcMetaExec(object):
    # -----------------------------------------------------------------------
    # CONSTRUCTOR
    # -----------------------------------------------------------------------

    def __init__(self, junos):
        """
        ~PRIVATE CLASS~
        creates an RPC meta-executor object bound to the provided
        ez-netconf :junos: object
        """
        self._junos = junos

    # -----------------------------------------------------------------------
    # get_config
    # -----------------------------------------------------------------------

    def get_config(
        self,
        filter_xml=None,
        options={},
        model=None,
        namespace=None,
        remove_ns=True,
        **kwargs
    ):
        """
        retrieve configuration from the Junos device

        .. code-block:: python

           dev.rpc.get_config()
           dev.rpc.get_config(filter_xml='<system><services/></system>')
           dev.rpc.get_config(filter_xml='system/services')
           dev.rpc.get_config(
               filter_xml=etree.XML('<system><services/></system>'),
               options={'format': 'json'})
           # to fetch junos as well as yang model configs
           dev.rpc.get_config(model=True)
           # openconfig yang example
           dev.rpc.get_config(filter_xml='bgp', model='openconfig')
           dev.rpc.get_config(filter_xml='<bgp><neighbors></neighbors></bgp>',
                            model='openconfig')
           # custom yang example
           dev.rpc.get_config(filter_xml='l2vpn', model='custom',
                        namespace="http://yang.juniper.net/customyang/l2vpn")
           # ietf yang example
           dev.rpc.get_config(filter_xml='interfaces', model='ietf')
           # ietf-softwire yang example
           dev.rpc.get_config(filter_xml='softwire-config', model='ietf',
                              namespace="urn:ietf:params:xml:ns:yang:ietf-softwire",
                              options={'format': 'json'})


        :filter_xml: fully XML formatted tag which defines what to retrieve,
                     when omitted the entire configuration is returned;
                     the following returns the device host-name configured
                     with "set system host-name"

        .. code-block:: python

           config = dev.rpc.get_config(filter_xml=etree.XML('''
               <configuration>
                   <system>
                       <host-name/>
                   </system>
               </configuration>'''))

        :options: is a dictionary of XML attributes to set within the
                  <get-configuration> RPC; the following returns the device
                  host-name either configured with "set system host-name"
                  and if unconfigured, the value inherited from
                  apply-group re0|re1, typical for multi-RE systems

        .. code-block:: python

           config = dev.rpc.get_config(filter_xml=etree.XML('''
                        <configuration>
                            <system>
                                <host-name/>
                            </system>
                        </configuration>'''),
                 options={'database':'committed','inherit':'inherit'})

        :param str model: Can provide yang model openconfig/custom/ietf. When
                model is True and filter_xml is None, xml is enclosed under
                <data> so that we get junos as well as other model
                configurations

        :param str namespace: User can have their own defined namespace in the
                custom yang models, In such cases they need to provide that
                namespace so that it can be used to fetch yang modeled configs

        :param bool remove_ns: remove namespaces, if value assigned is False,
                function will return xml with namespaces. The same xml
                returned can be loaded back to devices. This comes handy in
                case of yang based configs

        .. code-block:: python

           dev.rpc.get_config(filter_xml='bgp', model='openconfig',
                        remove_ns=False)
        """

        nmspaces = {
            "openconfig": "http://openconfig.net/yang/",
            "ietf": "urn:ietf:params:xml:ns:yang:ietf-",
        }

        rpc = E("get-configuration", options)

        if filter_xml is not None:
            if not isinstance(filter_xml, etree._Element):
                if re.search("^<.*>$", filter_xml):
                    filter_xml = etree.XML(filter_xml)
                else:
                    filter_data = None
                    for tag in filter_xml.split("/")[::-1]:
                        filter_data = (
                            E(tag) if filter_data is None else E(tag, filter_data)
                        )
                    filter_xml = filter_data
            # wrap the provided filter with toplevel <configuration> if
            # it does not already have one (not in case of yang model config)
            if (
                filter_xml.tag != "configuration"
                and model is None
                and namespace is None
            ):
                etree.SubElement(rpc, "configuration").append(filter_xml)
            else:
                if model is not None or namespace is not None:
                    if model == "custom" and namespace is None:
                        raise AttributeError(
                            'For "custom" model, ' 'explicitly provide "namespace"'
                        )
                    ns = namespace or (nmspaces.get(model.lower()) + filter_xml.tag)
                    filter_xml.attrib["xmlns"] = ns
                rpc.append(filter_xml)
        transform = self._junos.transform
        if remove_ns is False:
            self._junos.transform = lambda: JXML.strip_namespaces_prefix
        try:
            response = self._junos.execute(rpc, **kwargs)
        finally:
            self._junos.transform = transform
        # in case of model provided top level should be data
        # return response
        if model and filter_xml is None and options.get("format") != "json":
            response = response.getparent()
            response.tag = "data"
        return response

    # -----------------------------------------------------------------------
    # get
    # -----------------------------------------------------------------------

    def get(self, filter_select=None, ignore_warning=False, **kwargs):
        """
        Retrieve running configuration and device state information using
        <get> rpc

        .. code-block:: python

           dev.rpc.get()
           dev.rpc.get(ignore_warning=True)
           dev.rpc.get(filter_select='bgp') or dev.rpc.get('bgp')
           dev.rpc.get(filter_select='bgp/neighbors')
           dev.rpc.get("/bgp/neighbors/neighbor[neighbor-address='10.10.0.1']"
                       "/timers/state/hold-time")
           dev.rpc.get('mpls', ignore_warning=True)

        :param str filter_select:
          The select attribute will be treated as an XPath expression and
          used to filter the returned data.

        :param ignore_warning: A boolean, string or list of string.
          If the value is True, it will ignore all warnings regardless of the
          warning message. If the value is a string, it will ignore
          warning(s) if the message of each warning matches the string. If
          the value is a list of strings, ignore warning(s) if the message of
          each warning matches at least one of the strings in the list.

          For example::

            dev.rpc.get(ignore_warning=True)
            dev.rpc.get(ignore_warning='vrrp subsystem not running')
            dev.rpc.get(ignore_warning=['vrrp subsystem not running',
                                        'statement not found'])

          .. note::
            When the value of ignore_warning is a string, or list of strings,
            the string is actually used as a case-insensitive regular
            expression pattern. If the string contains only alpha-numeric
            characters, as shown in the above examples, this results in a
            case-insensitive substring match. However, any regular expression
            pattern supported by the re library may be used for more
            complicated match conditions.

        :returns: xml object
        """
        # junos only support filter type to be xpath
        filter_params = {"type": "xpath"}
        if filter_select is not None:
            filter_params["source"] = filter_select
        rpc = E("get", E("filter", filter_params))
        return self._junos.execute(rpc, ignore_warning=ignore_warning, **kwargs)

    # -----------------------------------------------------------------------
    # load_config
    # -----------------------------------------------------------------------

    def load_config(self, contents, ignore_warning=False, **options):
        """
        loads :contents: onto the Junos device, does not commit the change.

        :param ignore_warning: A boolean, string or list of string.
          If the value is True, it will ignore all warnings regardless of the
          warning message. If the value is a string, it will ignore
          warning(s) if the message of each warning matches the string. If
          the value is a list of strings, ignore warning(s) if the message of
          each warning matches at least one of the strings in the list.

          For example::

            dev.rpc.load_config(cnf, ignore_warning=True)
            dev.rpc.load_config(cnf,
                                ignore_warning='vrrp subsystem not running')
            dev.rpc.load_config(cnf,
                                ignore_warning=['vrrp subsystem not running',
                                                'statement not found'])
            dev.rpc.load_config(cnf, ignore_warning='statement not found')

          .. note::
            When the value of ignore_warning is a string, or list of strings,
            the string is actually used as a case-insensitive regular
            expression pattern. If the string contains only alpha-numeric
            characters, as shown in the above examples, this results in a
            case-insensitive substring match. However, any regular expression
            pattern supported by the re library may be used for more
            complicated match conditions.

        :options: is a dictionary of XML attributes to set within the
                  <load-configuration> RPC.

        The :contents: are interpreted by the :options: as follows:

        format='text' and action='set', then :contents: is a string containing
            a series of "set" commands

        format='text', then :contents: is a string containing Junos
            configuration in curly-brace/text format

        format='json', then :contents: is a string containing Junos
            configuration in json format

        url='path', then :contents: is a None

        <otherwise> :contents: is XML structure
        """
        rpc = E("load-configuration", options)

        if contents is None and "url" in options:
            pass
        elif ("action" in options) and (options["action"] == "set"):
            rpc.append(E("configuration-set", contents))
        elif ("action" in options) and (options["action"] == "patch"):
            rpc.append(E("configuration-patch", contents))
        elif ("format" in options) and (options["format"] == "text"):
            rpc.append(E("configuration-text", contents))
        elif ("format" in options) and (options["format"] == "json"):
            rpc.append(E("configuration-json", contents))
        else:
            # otherwise, it's just XML Element
            if contents.tag != "configuration":
                etree.SubElement(rpc, "configuration").append(contents)
            else:
                rpc.append(contents)

        return self._junos.execute(rpc, ignore_warning=ignore_warning)

    # -----------------------------------------------------------------------
    # cli
    # -----------------------------------------------------------------------

    def cli(self, command, format="text", normalize=False):
        rpc = E("command", command)
        if format.lower() in ["text", "json"]:
            rpc.attrib["format"] = format
        return self._junos.execute(rpc, normalize=normalize)

    # -----------------------------------------------------------------------
    # method missing
    # -----------------------------------------------------------------------

    def __getattr__(self, rpc_cmd_name):
        """
        metaprograms a function to execute the :rpc_cmd_name:

        the caller will be passing (*vargs, **kvargs) on
        execution of the meta function; these are the specific
        rpc command arguments(**kvargs) and options bound
        as XML attributes (*vargs)
        """

        rpc_cmd = re.sub("_", "-", rpc_cmd_name)

        def _exec_rpc(*vargs, **kvargs):
            # create the rpc as XML command
            rpc = etree.Element(rpc_cmd)

            # Gather decorator keywords into dec_args and remove from kvargs
            dec_arg_keywords = [
                "dev_timeout",
                "normalize",
                "ignore_warning",
                "filter_xml",
            ]
            dec_args = {}
            for keyword in dec_arg_keywords:
                if keyword in kvargs:
                    dec_args[keyword] = kvargs.pop(keyword)

            # kvargs are the command parameter/values
            if kvargs:
                for arg_name, arg_value in kvargs.items():
                    arg_name = re.sub("_", "-", arg_name)
                    if not isinstance(arg_value, (tuple, list)):
                        arg_value = [arg_value]
                    for a in arg_value:
                        if not isinstance(
                            a,
                            (bool, str, unicode) if sys.version < "3" else (bool, str),
                        ):
                            raise TypeError(
                                "The value %s for argument %s"
                                " is of %s. Argument "
                                "values must be a string, "
                                "boolean, or list/tuple of "
                                "strings and booleans." % (a, arg_name, str(type(a)))
                            )
                        if a is not False:
                            arg = etree.SubElement(rpc, arg_name)
                        if not isinstance(a, bool):
                            arg.text = a

            # vargs[0] is a dict, command options like format='text'
            if vargs:
                for k, v in vargs[0].items():
                    if v is not True:
                        rpc.attrib[k] = v

            # now invoke the command against the
            # associated :junos: device and return
            # the results per :junos:execute()
            return self._junos.execute(rpc, **dec_args)

        # metabind help() and the function name to the :rpc_cmd_name:
        # provided by the caller ... that's about all we can do, yo!

        _exec_rpc.__doc__ = rpc_cmd
        _exec_rpc.__name__ = rpc_cmd_name

        # return the metafunction that the caller will in-turn invoke
        return _exec_rpc

    # -----------------------------------------------------------------------
    # callable
    # -----------------------------------------------------------------------

    def __call__(self, rpc_cmd, **kvargs):
        """
        callable will execute the provided :rpc_cmd: against the
        attached :junos: object and return the RPC response per
        :junos:execute()

        kvargs is simply passed 'as-is' to :junos:execute()
        """
        return self._junos.execute(rpc_cmd, **kvargs)



def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "2RE": "A boolean indicating if the device has more than one "
        "Routing Engine installed.",
        "master": "On a single chassis/node system, a string value of "
        "'RE0' or 'RE1' indicating which RE is master. On a "
        "multi-chassis or multi-node system, the value is a "
        "list of these strings indicating whether RE0 or RE1 "
        "is master. There is one entry in the list for each "
        "chassis/node in the system.",
        "RE0": "A dictionary with information about RE0 (if present). The "
        "keys of the dictionary are: mastership_state, status, "
        "model, up_time, and last_reboot_reason.",
        "RE1": "A dictionary with information about RE1 (if present). The "
        "keys of the dictionary are: mastership_state, status, "
        "model, up_time, and last_reboot_reason.",
        "re_info": "A three-level dictionary with information about "
        "the Routing Engines in the device. The first-level "
        "key is the chassis or node name. The second-level key "
        "is the slot number, the third-level keys are: "
        "mastership_state, status, model, and "
        "last_reboot_reason. A first-level key with a value "
        "of 'default' will always be present and represents "
        "the first chassis/node of the system (Note: the first "
        "chasis/node of the system is not necessarily the "
        "'master' node in a VC.) A second-level key with a "
        "value of 'default' will always be present "
        "for the default chassis/node and represents the "
        "first Routing Engine on the first node/chassis. "
        "(Note: the first RE of a chassis/node is not "
        "necessarily the 'master' RE of the chassis/node. See "
        "the RE_master fact for info on the 'master' RE of "
        "each chassis/node.)",
        "re_master": "A dictionary indicating which RE slot is master for "
        "each chassis/node in the system. The dictionary key "
        "is the chassis or node name. A key with a value "
        "of 'default' will always be present and represents "
        "the first node/chassis of the system. (Note: the "
        "first chassis/node of the system is not necessarily "
        "the 'master' node in a VC. See the vc_master fact "
        "to determine which chassis/node is the master of "
        "a VC.)",
    }


def get_facts(device):
    """
    Gathers facts from the <get-route-engine-information/> RPC.
    """
    multi_re = None
    master = None
    RE0 = None
    RE1 = None
    re_info = None
    re_master = None

    rsp = device.rpc.get_route_engine_information(normalize=True)
    re_list = rsp.findall(".//route-engine")
    if len(re_list) > 1:
        multi_re = True
    else:
        multi_re = False

    first_node = None
    first_slot = None
    master_list = []
    node_masters = {}
    for current_re in re_list:
        node = current_re.findtext("../../re-name", "default")
        slot = current_re.findtext("slot", "0")
        info = {
            "mastership_state": current_re.findtext("mastership-state", "master"),
            "status": current_re.findtext("status"),
            "model": current_re.findtext("model"),
            "last_reboot_reason": current_re.findtext("last-reboot-reason"),
            # This key is only returned in the RE0 and RE1 facts in order
            # to maintain backwards compatibility with the old fact
            # gathering system. Since the up_time value changes, it's not
            # really a "fact" and is therefore omitted from the new re_info
            # fact.
            "up_time": current_re.findtext("up-time"),
        }
        if first_node is None:
            first_node = node
            first_slot = slot
        if node == first_node:
            if slot == "0" and RE0 is None:
                # Copy the dictionary
                RE0 = dict(info)
            if slot == "1" and RE1 is None:
                # Copy the dictionary
                RE1 = dict(info)
        # Don't want the up_time key in the new re_info fact.
        if "up_time" in info:
            del info["up_time"]
        if re_info is None:
            re_info = {}
        if node not in re_info:
            re_info[node] = {}
        # Save with second-level key as a string.
        re_info[node][slot] = info
        # If it's a master RE, then save in node_masters and master_list
        if "mastership_state" in info and info["mastership_state"].lower().startswith(
            "master"
        ):
            node_masters[node] = slot
            master_list.append("RE" + slot)
    # Fill in the 'default' first-level key if multi-chassis/node system
    if first_node is not None and first_node != "default":
        re_info["default"] = re_info[first_node]
        if first_node in node_masters:
            node_masters["default"] = node_masters[first_node]
    # Fill in the 'default' second-level key if at least one RE was found.
    if first_slot is not None:
        re_info["default"]["default"] = re_info["default"][first_slot]

    # Set the 'master' fact to a string or list based on the number of members.
    master_list_len = len(master_list)
    if master_list_len == 1:
        master = master_list[0]
    elif master_list_len > 1:
        master = master_list

    # If any info in the node_masters dict, then return it as fact 're_master'
    if node_masters:
        re_master = node_masters

    return {
        "2RE": multi_re,
        "master": master,
        "RE0": RE0,
        "RE1": RE1,
        "re_info": re_info,
        "re_master": re_master,
    }



from jnpr.junos.exception import ConnectNotMasterError
from jnpr.junos.exception import RpcError


def facts_chassis(junos, facts):
    """
    The following facts are assigned:
      facts['2RE'] : designates if the device can support two RE, not that it
                     has them
      facts['RE_hw_mi'] : designates if the device is
                          multi-instance-routing-engine
      facts['model'] : product model
      facts['serialnumber'] : serial number

    NOTES:
        (1) if in a 2RE system, this routine will only load the information
            from the first chassis item.
        (2) hostname, domain, and fqdn are retrieved from configuration data;
            inherited configs are checked.
    """
    # Set default values.
    facts["2RE"] = False
    facts["RE_hw_mi"] = False
    facts["model"] = "UNKNOWN"
    facts["serialnumber"] = "UNKNOWN"

    rsp = junos.rpc.get_chassis_inventory()
    if rsp.tag == "error":
        raise RuntimeError()

    if rsp.tag == "output":
        # this means that there was an error; due to the
        # fact that this connection is not on the master
        # @@@ need to validate on VC-member
        raise ConnectNotMasterError(junos)

    if rsp.tag == "multi-routing-engine-results":
        facts["2RE"] = True
        facts["RE_hw_mi"] = True
    else:
        facts["2RE"] = False

    facts["model"] = rsp.findtext(".//chassis[1]/description", "UNKNOWN")
    facts["serialnumber"] = (
        rsp.findtext(".//chassis[1]/serial-number")
        or rsp.findtext('.//chassis-module[name="Backplane"]/serial-number')
        or rsp.findtext('.//chassis-module[name="Midplane"]/serial-number', "UNKNOWN")
    )

    if facts["model"] == "UNKNOWN" or facts["serialnumber"] == "UNKNOWN":
        raise RpcError()



# 3rd-party
from lxml.builder import E

# local
from jnpr.junos.cfg.resource import Resource
from jnpr.junos import JXML
from jnpr.junos.cfg.phyport.base import PhyPortBase


class PhyPortClassic(PhyPortBase):
    # -----------------------------------------------------------------------
    # XML readers
    # -----------------------------------------------------------------------

    def _xml_to_py(self, has_xml, has_py):
        PhyPortBase._xml_to_py(self, has_xml, has_py)

        Resource.copyifexists(has_xml, "speed", has_py)
        Resource.copyifexists(has_xml, "link-mode", has_py, "duplex")
        if has_xml.find("gigether-options/loopback") is not None:
            has_py["loopback"] = True
        has_py["$unit_count"] = len(has_xml.findall("unit"))

        # normalizers
        if "duplex" in has_py:
            PhyPortBase._set_invert(has_py, "duplex", self.PORT_DUPLEX)

    # -----------------------------------------------------------------------
    # XML writers
    # -----------------------------------------------------------------------

    def _xml_change_speed(self, xml):
        Resource.xml_set_or_delete(xml, "speed", self.speed)
        return True

    def _xml_change_duplex(self, xml):
        value = self.PORT_DUPLEX.get(self.duplex)
        Resource.xml_set_or_delete(xml, "link-mode", value)
        return True

    def _xml_change_loopback(self, xml):
        opts = E("gigether-options")
        opts.append(Resource.xmltag_set_or_del("loopback", self.loopback))
        xml.append(opts)
        return True



from jnpr.junos.exception import RpcError
import re
from lxml import etree


def _get_vmhost_version_information(device):
    multi_re = False
    try:
        rsp = device.rpc.get_route_engine_information(normalize=True)
        re_list = rsp.findall(".//route-engine")
        if len(re_list) > 1:
            multi_re = True
        else:
            multi_re = False
    except RpcError as err:
        raise RpcError()

    if multi_re == True:
        try:
            return device.rpc.cli(
                "show vmhost version invoke-on all-routing-engines",
                format="xml",
                normalize=True,
            )
        except RpcError as err:
            raise RpcError()
    else:
        try:
            return device.rpc.cli("show vmhost version", format="xml", normalize=True)
        except RpcError as err:
            raise RpcError()


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "vmhost": "A boolean indicating if the device is vmhost.",
        "vmhost_info": "A dictionary indicating  vmhost RE partition JUNOS versions.",
    }


def get_facts(device):
    """
    Gathers facts from the sysctl command.
    """
    SYSCTL_VMHOST_MODE = "sysctl -n hw.re.vmhost_mode"
    vmhost = None
    vmhost_info = {}
    vm_ver_rsp = None
    vmhost_current_root_set = None
    vmhost_set_junos_version_set_p = None
    vmhost_set_junos_version_set_b = None
    re_name = None

    if device.facts["_is_linux"]:
        vmhost = False
    else:
        try:
            rsp = device.rpc.request_shell_execute(command=SYSCTL_VMHOST_MODE)
            if rsp.tag == "rpc-error":
                raise RpcError()
            result = re.sub("<[^<]+>", "", etree.tostring(rsp).decode())
            if result.strip() == "1":
                vmhost = True
            else:
                vmhost = False
        except RpcError:
            pass

    if vmhost:
        rsp = _get_vmhost_version_information(device)
        # PR 1510446 fix for show vmhost version rpc supports form 22.2R3
        if device.facts["version"] >= "22.2R3":
            if rsp.tag == "vmhost-version-information":
                vm_ver_rsp = [rsp]
            else:
                vm_ver_rsp = rsp.findall(".//vmhost-version-information")

            for re_vm_ver_info in vm_ver_rsp:
                re_name = re_vm_ver_info.findtext("../re-name", "re0")
                vmhost_current_root_set = re_vm_ver_info.findtext("./current-root-set")
                vmhost_set_junos_version_set_p = re_vm_ver_info.findtext(
                    "./set-disk-info[set-disk-name = 'set p']/set-junos-version"
                )
                vmhost_set_junos_version_set_b = re_vm_ver_info.findtext(
                    "./set-disk-info[set-disk-name = 'set b']/set-junos-version"
                )

                vmhost_info[re_name] = {
                    "vmhost_current_root_set": vmhost_current_root_set,
                    "vmhost_version_set_p": vmhost_set_junos_version_set_p,
                    "vmhost_version_set_b": vmhost_set_junos_version_set_b,
                }
        else:
            pass

    return {
        "vmhost": vmhost,
        "vmhost_info": vmhost_info,
    }



import traceback
import sys
import logging
import warnings
import socket

# 3rd-party packages
from ncclient.devices.junos import JunosDeviceHandler
from lxml import etree
from jnpr.junos.transport.tty_telnet import Telnet
from jnpr.junos.transport.tty_serial import Serial
from jnpr.junos.transport.tty_ssh import SSH
from ncclient.xml_ import NCElement
from jnpr.junos.device import _Connection

# local modules
from jnpr.junos.rpcmeta import _RpcMetaExec
from jnpr.junos.factcache import _FactCache
from jnpr.junos import jxml as JXML
from jnpr.junos.ofacts import *
from jnpr.junos.decorators import ignoreWarnDecorator
from jnpr.junos.device import _Jinja2ldr

logger = logging.getLogger("jnpr.junos.console")


class Console(_Connection):
    def __init__(self, **kvargs):
        """
        NoobDevice object constructor.

        :param str host:
            **REQUIRED** host-name or ipaddress of target device

        :param str user:
            *OPTIONAL* login user-name, uses root if not provided

        :param str passwd:
            *OPTIONAL* in console connection for device at zeroized state
            password is not required

        :param int port:
            *OPTIONAL*  port, defaults to '23' for telnet mode and
            '/dev/ttyUSB0' for serial.

        :param int baud:
            *OPTIONAL*  baud, default baud rate is 9600

        :param str mode:
            *OPTIONAL*  mode, mode of connection (telnet/serial)
            default is telnet

        :param int timeout:
            *OPTIONAL*  timeout, default is 0.5

        :param int attempts:
            *OPTIONAL*  attempts, default is 10

        :param str ssh_config:
            *OPTIONAL* The path to the SSH configuration file.
            This can be used to load SSH information from a configuration file.
            By default ~/.ssh/config is queried it will be used by SCP class.
            So its assumed ssh is enabled by the time we use SCP functionality.

        :param bool gather_facts:
            *OPTIONAL* Defaults to ``False``. If ``False`` and old-style fact
            gathering is in use then facts are not gathered on call to
            :meth:`open`. This argument is a no-op when new-style fact
            gathering is in use (the default.)

        :param str fact_style:
            *OPTIONAL*  The style of fact gathering to use. Valid values are:
            'new', 'old', or 'both'. The default is 'new'. The value 'both' is
            only present for debugging purposes. It will be removed in a future
            release. The value 'old' is only present to workaround bugs in
            new-style fact gathering. It will be removed in a future release.

        :param bool console_has_banner:
            *OPTIONAL* default is ``False``.  If ``False`` then in case of a
            hung state, <close-session/> rpc is sent to the console.
            If ``True``, after sleep(5), a new-line is sent

        """

        # ----------------------------------------
        # setup instance connection/open variables
        # ----------------------------------------

        self._tty = None
        self._ofacts = {}
        self.connected = False
        self._skip_logout = False
        self.results = dict(changed=False, failed=False, errmsg=None)

        # hostname is not required in serial mode connection
        self._hostname = kvargs.get("host")
        self._auth_user = kvargs.get("user", "root")
        self._conf_auth_user = None
        self._conf_ssh_private_key_file = None
        self._auth_password = kvargs.get("password", "") or kvargs.get("passwd", "")
        self.cs_user = kvargs.get("cs_user")
        self.cs_passwd = kvargs.get("cs_passwd")
        self._port = kvargs.get("port", "22" if self.cs_user else "23")
        self._mode = kvargs.get("mode", None if self.cs_user else "telnet")
        self._baud = kvargs.get("baud", "9600")
        if self._hostname:
            self._ssh_config = kvargs.get("ssh_config")
            self._sshconf_lkup()
        self._ssh_private_key_file = (
            kvargs.get("ssh_private_key_file") or self._conf_ssh_private_key_file
        )
        self._timeout = kvargs.get("timeout", "0.5")
        self._normalize = kvargs.get("normalize", False)
        self._attempts = kvargs.get("attempts", 10)
        self._gather_facts = kvargs.get("gather_facts", False)
        self._fact_style = kvargs.get("fact_style", "new")
        self._use_filter = False
        self._huge_tree = kvargs.get("huge_tree", False)
        if self._fact_style != "new":
            warnings.warn(
                "fact-style %s will be removed in "
                "a future release." % (self._fact_style),
                RuntimeWarning,
            )
        self.console_has_banner = kvargs.get("console_has_banner", False)
        self.rpc = _RpcMetaExec(self)
        self._manages = []
        self.junos_dev_handler = JunosDeviceHandler(
            device_params={"name": "junos", "local": False}
        )
        self._conn = None
        self._j2ldr = _Jinja2ldr
        if self._fact_style == "old":
            self.facts = self.ofacts
        else:
            self.facts = _FactCache(self)

    @property
    def timeout(self):
        """
        :returns: current console connection timeout value (int) in seconds.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        """
        Used to change the console connection timeout value (default=0.5 sec).

        :param int value:
            New timeout value in seconds
        """
        self._timeout = value

    @property
    def transform(self):
        """
        :returns: the current RPC XML Transformation.
        """
        return self.junos_dev_handler.transform_reply

    @transform.setter
    def transform(self, func):
        """
        Used to change the RPC XML Transformation.

        :param lambda value:
            New transform lambda
        """
        self.junos_dev_handler.transform_reply = func

    def open(self, *vargs, **kvargs):
        """
        Opens a connection to the device using existing login/auth
        information.

        :param bool gather_facts:
            If set to ``True``/``False`` will override the device
            instance value for only this open process
        """

        # ---------------------------------------------------------------
        # validate device hostname or IP address
        # ---------------------------------------------------------------
        if (
            (self._mode and self._mode.upper() == "TELNET") or self.cs_user is not None
        ) and self._hostname is None:
            self.results["failed"] = True
            self.results["errmsg"] = "ERROR: Device hostname/IP not specified !!!"
            return self.results

        # ---------------------------------------------------------------
        # validate console server and password. Password-less connection
        # is not supported
        # ---------------------------------------------------------------
        if self.cs_user is not None and self.cs_passwd is None:
            self.results["failed"] = True
            self.results["errmsg"] = (
                "ERROR: Console SSH, Password-less connection is " "not supported !!!"
            )
            logger.error(self.results["errmsg"])
            return self.results

        # --------------------
        # login to the CONSOLE
        # --------------------
        try:
            self._tty_login()
        except RuntimeError as err:
            logger.error("ERROR:  {}:{}\n".format("login", str(err)))
            logger.error(
                "\nComplete traceback message: {}".format(traceback.format_exc())
            )
            raise err
        except Exception as ex:
            logger.error("Exception occurred: {}:{}\n".format("login", str(ex)))
            raise ex
        self.connected = True

        self._nc_transform = self.transform
        self._norm_transform = lambda: JXML.normalize_xslt.encode("UTF-8")

        # normalize argument to open() overrides normalize argument value
        # to __init__(). Save value to self._normalize where it is used by
        # normalizeDecorator()
        self._normalize = kvargs.get("normalize", self._normalize)
        if self._normalize is True:
            self.transform = self._norm_transform

        gather_facts = kvargs.get("gather_facts", self._gather_facts)
        if gather_facts is True:
            logger.info("facts: retrieving device facts...")
            self.facts_refresh()
            self.results["facts"] = self.facts
        self._conn = self._tty
        return self

    def close(self, skip_logout=False):
        """
        Closes the connection to the device.
        """
        if skip_logout is False and self.connected is True:
            try:
                self._tty_logout()
            except socket.error as err:
                # if err contains "Connection reset by peer" connection to the
                # device got closed
                if "Connection reset by peer" not in str(err):
                    raise err
            except EOFError as err:
                if "telnet connection closed" not in str(err):
                    raise err
            except Exception as err:
                logger.error("ERROR {}:{}\n".format("logout", str(err)))
                raise err
            self.connected = False
        elif self.connected is True:
            try:
                self._tty._tty_close()
            except Exception as err:
                logger.error("ERROR {}:{}\n".format("close", str(err)))
                logger.error(
                    "\nComplete traceback message: {}".format(traceback.format_exc())
                )
                raise err
            self.connected = False

    @ignoreWarnDecorator
    def _rpc_reply(self, rpc_cmd_e, *args, **kwargs):
        encode = None if sys.version < "3" else "unicode"
        rpc_cmd = (
            etree.tostring(rpc_cmd_e, encoding=encode)
            if isinstance(rpc_cmd_e, etree._Element)
            else rpc_cmd_e
        )
        reply = self._tty.nc.rpc(rpc_cmd)
        rpc_rsp_e = NCElement(
            reply, self.junos_dev_handler.transform_reply(), self._huge_tree
        )._NCElement__doc
        return rpc_rsp_e

    # -------------------------------------------------------------------------
    # LOGIN/LOGOUT
    # -------------------------------------------------------------------------

    def _tty_login(self):
        tty_args = dict()
        tty_args["user"] = self._auth_user
        tty_args["passwd"] = self._auth_password
        tty_args["timeout"] = float(self._timeout)
        tty_args["attempts"] = int(self._attempts)
        tty_args["baud"] = self._baud
        tty_args["huge_tree"] = self._huge_tree
        if self._mode and self._mode.upper() == "TELNET":
            tty_args["host"] = self._hostname
            tty_args["port"] = self._port
            tty_args["console_has_banner"] = self.console_has_banner
            self.console = ("telnet", self._hostname, self.port)
            self._tty = Telnet(**tty_args)
        elif self.cs_user is not None:
            tty_args["cs_user"] = self.cs_user
            tty_args["cs_passwd"] = self.cs_passwd
            tty_args["host"] = self._hostname
            tty_args["port"] = self._port
            tty_args["console_has_banner"] = self.console_has_banner
            tty_args["ssh_private_key_file"] = self._ssh_private_key_file
            self.console = ("ssh", self._hostname, self.port)
            self._tty = SSH(**tty_args)
        elif self._mode.upper() == "SERIAL":
            tty_args["port"] = self._port
            self.console = ("serial", self._port)
            self._tty = Serial(**tty_args)
        else:
            logger.error("Mode should be either telnet or serial")
            raise AttributeError("Mode to be telnet/serial")

        self._tty.login()

    def _tty_logout(self):
        self._tty.logout()

    def zeroize(self):
        """perform device ZEROIZE actions"""
        logger.info("zeroize : ZEROIZE device, rebooting")
        self._tty.nc.zeroize()
        self._skip_logout = True
        self.results["changed"] = True

    # -----------------------------------------------------------------------
    # Context Manager
    # -----------------------------------------------------------------------

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.connected:
            self.close()



from jnpr.junos.cfg import Resource
from jnpr.junos.cfg.phyport.classic import PhyPortClassic
from jnpr.junos.cfg.phyport.switch import PhyPortSwitch

__all__ = ["PhyPort"]


class PhyPort(object):
    def __new__(cls, dev, name=None):
        supercls = {
            "CLASSIC": PhyPortClassic,
            "SWITCH": PhyPortSwitch,
        }.get(dev.facts["ifd_style"])
        newcls = type(cls.__name__, (supercls,), {})
        return newcls(dev, name)



from lxml import etree
from jnpr.junos.exception import PermissionError
from jnpr.junos.utils.fs import FS


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "domain": "The domain name configured at the [edit system "
        "domain-name] configuration hierarchy.",
        "fqdn": "The device's hostname + domain",
    }


def get_facts(device):
    """
    Gathers domain facts from the configuration or /etc/resolv.conf.
    """
    domain_config = """
        <configuration>
            <system>
                <domain-name/>
             </system>
        </configuration>
    """
    domain = None
    fqdn = None
    # Try to read the domain-name from the config.
    # This might fail due to lack of permissions.
    try:
        rsp = device.rpc.get_config(
            filter_xml=etree.XML(domain_config),
            options={
                "database": "committed",
                "inherit": "inherit",
                "commit-scripts": "apply",
            },
        )
        domain = rsp.findtext(".//domain-name")
    # Ignore if user can't view the configuration.
    except PermissionError:
        pass

    # Try to read the domain from the resolv.conf file. This only requires
    # view permissions.
    if domain is None:
        fs = FS(device)
        file_content = fs.cat("/etc/resolv.conf") or fs.cat("/var/etc/resolv.conf")
        words = file_content.split() if file_content is not None else []
        if "domain" in words:
            idx = words.index("domain") + 1
            domain = words[idx]

    # Set the fqdn
    fqdn = device.facts["hostname"]
    if fqdn is not None and domain is not None:
        fqdn = fqdn + "." + domain

    return {
        "domain": domain,
        "fqdn": fqdn,
    }



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.user_ssh_key import UserSSHKey


class User(Resource):
    """
    [edit system login user <name>]

    Resource name: str
      <name> is the user login name

    Manages resources:
      sshkey, UserSSHKey
    """

    PROPERTIES = [
        "uid",
        "fullname",  # the full-name field
        "userclass",  # user class
        "password",  # write-only clear-text password, will get crypt'd
        "$password",  # read-only crypt'd password
        "$sshkeys",  # read-only names of ssh-keys
    ]

    MANAGES = {"sshkey": UserSSHKey}

    # -----------------------------------------------------------------------
    # XML readers
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.system(E.login(E.user(E.name(self._name))))

    def _xml_at_res(self, xml):
        return xml.find(".//user")

    def _xml_to_py(self, has_xml, has_py):
        Resource._r_has_xml_status(has_xml, has_py)

        has_py["userclass"] = has_xml.findtext("class")

        Resource.copyifexists(has_xml, "full-name", has_py, "fullname")

        Resource.copyifexists(has_xml, "uid", has_py)
        if "uid" in has_py:
            has_py["uid"] = int(has_py["uid"])

        auth = has_xml.find("authentication")
        if auth is not None:
            # plain-text password
            Resource.copyifexists(auth, "encrypted-password", has_py, "$password")

            # ssh-keys
            sshkeys = auth.xpath("ssh-rsa | ssh-dsa")
            if sshkeys is not None:
                has_py["$sshkeys"] = [
                    (sshkey.tag, sshkey.findtext("name").strip()) for sshkey in sshkeys
                ]

    # -----------------------------------------------------------------------
    # XML property writers
    # -----------------------------------------------------------------------

    def _xml_change_fullname(self, xml):
        xml.append(E("full-name", self["fullname"]))
        return True

    def _xml_change_userclass(self, xml):
        xml.append(E("class", self["userclass"]))
        return True

    def _xml_change_password(self, xml):
        xml.append(E.authentication(E("plain-text-password-value", self["password"])))
        return True

    def _xml_change_uid(self, xml):
        xml.append(E.uid(str(self["uid"])))
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        get = E.system(E.login(E.user(JXML.NAMES_ONLY)))
        got = self.R.get_config(get)
        self._rlist = [name.text for name in got.xpath(".//user/name")]



class CMDView(object):
    """
    View is the base-class that makes extracting values from XML
    data appear as objects with attributes.
    """

    KEY = "name"
    KEY_ITEMS = []
    COLUMNS = {}
    FILTERS = None
    FIELDS = {}
    REGEX = {}
    EXISTS = {}
    GROUPS = None
    TITLE = None
    EVAL = {}

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def T(self):
        """return the Table instance for the View"""
        return self._table

    @property
    def D(self):
        """return the Device instance for this View"""
        return self.T.D

    @property
    def xml(self):
        """returns the XML associated to the item"""
        return self._xml

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def keys(self):
        """list of view keys, i.e. field names"""
        return self.FIELDS.keys()

    def values(self):
        """list of view values"""
        return [getattr(self, field) for field in self.keys()]

    def items(self):
        """list of tuple(key,value)"""
        return zip(self.keys(), self.values())

    def asview(self, view_cls):
        """create a new View object for this item"""
        return view_cls(self._table, self._xml)

    # -------------------------------------------------------------------------
    # OVERLOADS
    # -------------------------------------------------------------------------

    def __repr__(self):
        """returns the name of the View with the associate item name"""
        return "%s:%s" % (self.__class__.__name__, self.name)

    def __getitem__(self, name):
        """
        allow the caller to extract field values using :obj['name']:
        the same way they would do :obj.name:
        """
        return getattr(self, name)



import warnings
from pprint import pformat

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

import jnpr.junos.facts
from jnpr.junos.facts import __doc__ as facts_doc
import jnpr.junos.exception


class _FactCache(MutableMapping):
    """
    A dictionary-like object which performs on-demand fact gathering.

    This class should not be used directly. An instance of this class is
    available as the :attr:`facts` attribute of a Device object.

    **Dictionary magic methods:**
      * :meth:`__getitem__`: Gets the value of a given key in the dict.
      * :meth:`__delitem__`: Called when a key is deleted from the dict.
      * :meth:`__setitem__`: Called when a key is set on the dict.
      * :meth:`__iter__`: Called when iterating over the keys of the dict.
      * :meth:`__len__`: Called when getting the length of the dict.
      * :meth:`__repr__`: Called when representing the dict as a string.

    **Additional methods:**
      * :meth:`_refresh`: Refreshes the fact cache.
    """

    def __init__(self, device):
        """
        _FactCache object constructor.

        :param device: The device object for which fact gathering will be
          performed.
        """
        self._device = device
        self._cache = dict()
        self._call_stack = list()
        self._callbacks = jnpr.junos.facts._callbacks
        self._exception_on_failure = False
        self._warnings_on_failure = False
        self._should_warn = False

    def __getitem__(self, key):
        """
        Return the value of a particular key in the dictionary.

        If the fact has already been cached, the value is simply returned from
        the cache. If the value has not been cached, then the appropriate
        callback function is invoked to gather the fact from the device. The
        value is cached, and then returned.

        If _warnings_on_failure is True, then a warning is logged if there is
        an error gathering a fact from the device.

        :param key: The key who's value is returned.

        :returns value: The value of the key fact. If key is a known fact, but
            there is an error gathering the fact, then the value None is
            returned.

        :raises KeyError:
            When key is not a known fact (there is no callback present to
            gather the fact.)

        :raises jnpr.junos.exception.FactLoopError:
            When there is a loop attempting to gather the fact.

        :raises other exceptions as defined by the fact gathering modules:
            When an error is encountered and _exception_on_failure is True.
        """
        if key not in self._callbacks:
            # Not a fact that we know how to provide.
            raise KeyError(
                "%s: There is no function to gather the %s fact" % (key, key)
            )
        if key not in self._cache:
            # A known fact, but not yet cached. Go get it and cache it.
            if self._callbacks[key] in self._call_stack:
                raise jnpr.junos.exception.FactLoopError(
                    "A loop was detected while gathering the %s fact. The %s "
                    "module has already been called. Please report this error."
                    % (key, self._callbacks[key].__module__)
                )
            else:
                # Add the callback we are about to invoke to the _call_stack in
                # order to detect loops in fact gathering.
                self._call_stack.append(self._callbacks[key])
            try:
                # Invoke the callback
                new_facts = self._callbacks[key](self._device)
            except jnpr.junos.exception.FactLoopError:
                raise
            except Exception:
                # An exception was raised. No facts were returned.
                # Raise the exception to the user?
                if self._exception_on_failure:
                    raise
                # Warn the user?
                if self._warnings_on_failure:
                    self._should_warn = True
                # Set all of the facts which should have been returned
                # by this callback to the default value of None.
                for new_key in self._callbacks:
                    if self._callbacks[key] is self._callbacks[new_key]:
                        self._cache[new_key] = None
            else:
                # No exception
                for new_key in new_facts:
                    if (
                        new_key not in self._callbacks
                        or self._callbacks[key] is not self._callbacks[new_key]
                    ):
                        # The callback returned a fact it didn't advertise
                        raise RuntimeError(
                            "The %s module returned the %s "
                            "fact, but does not list %s as a "
                            "provided fact. Please report this "
                            "error."
                            % (self._callbacks[key].__module__, new_key, new_key)
                        )
                    else:
                        # Cache the returned fact
                        self._cache[new_key] = new_facts[new_key]
            finally:
                # Always pop the current callback from _call_stack,
                # regardless of whether or not an exception was raised.
                self._call_stack.pop()
        if key in self._cache:
            # key fact is cached. Return it.
            if self._device._fact_style == "both":
                # Compare old and new-style values.
                if key in self._device._ofacts:
                    # Skip key comparisons for certain keys.
                    #
                    # The old facts gathering code has an up_time key.
                    # The new facts gathering code maintains this key for RE0
                    # and RE1 facts, but it's not comparable (because it
                    # depends on when the fact was gathered and is therefore
                    # not really a "fact".) The new re_info fact omits the
                    # up_time field for this reason.
                    #
                    # The old facts gathering code didn't return a correct
                    # value for the master fact when the system was a VC.
                    # The new fact gathering code still returns the master fact
                    # but returns a correct value for VCs. It also returns a
                    # new re_master fact which is much more useful.
                    if key not in ["RE0", "RE1", "master"]:
                        if self._cache[key] != self._device._ofacts[key]:
                            warnings.warn(
                                "New and old-style facts do not "
                                "match for the %s fact.\n"
                                "    New-style value: %s\n"
                                "    Old-style value: %s\n"
                                % (key, self._cache[key], self._device._ofacts[key]),
                                RuntimeWarning,
                            )
            return self._cache[key]
        else:
            # key fact was not returned by callback
            raise RuntimeError(
                "The %s module claims to provide the %s "
                "fact, but failed to return it. Please report "
                "this error." % (self._callbacks[key].__module__, key)
            )

    def __delitem__(self, key):
        """
        Facts are read-only. Don't allow deleting an item.
        """
        raise RuntimeError("facts are read-only!")

    def __setitem__(self, key, value):
        """
        Facts are read-only. Don't allow setting an item.
        """
        raise RuntimeError("facts are read-only!")

    def __iter__(self):
        """
        An iterator of known facts.

        :returns iterator: of all of the 'non-hidden' facts we know how to
        gather, regardless of whether or not they've already been cached. Fact
        names which are hidden start with an underscore and are not returned.
        """
        callbacks = {}
        for key in self._callbacks:
            if not key.startswith("_"):
                callbacks[key] = self._callbacks[key]
        return iter(callbacks)

    def __len__(self):
        """
        The length of all known facts.

        :returns length: of all of the facts we know how to gather,
        regardless of whether or not they've already been cached.
        """
        return len(self._callbacks)

    def __str__(self):
        """
        A string representation of the facts dictionary.

        :returns string: a string representation of the dictionary.
          Because this returns the value of every fact, it has the
          side-effect of causing any ungathered facts to be gathered and then
          cached.
        """
        string = ""
        for key in sorted(self):
            if not key.startswith("_"):
                current = "'%s': %s" % (key, repr(self.get(key)))
                if string:
                    string = ", ".join([string, current])
                else:
                    string = current
        return "{" + string + "}"

    def __repr__(self):
        """
        A formated string representation of the facts dictionary.

        :returns string: a formated string representation of the dictionary.
          Because this returns the value of every fact, it has the
          side-effect of causing any ungathered facts to be gathered and then
          cached.
        """
        return pformat(dict(self))

    def _refresh(
        self, exception_on_failure=False, warnings_on_failure=False, keys=None
    ):
        """
        Empty the cache to force a refresh of one or more facts.

        Empties the fact gathering cache for all keys (if keys == None) or a
        set of keys. This causes the fact to be gathered and cached upon next
        access. If either eception_on_failure or warnings_on_failure is true,
        then all facts are accessed by getting the string representation of the
        facts. This causes all facts to immediately be populated so that any
        exceptions or warnings are generated during the call to _refresh().

        :param exception_on_failure: A boolean which indicates if an exception
          should be raised upon a failure gathering facts.

        :param warnings_on_failure: A boolean which indicates if an warning
          should be logged upon a failure gathering facts.

        :param keys: A single key as a string, or an iterable of keys (such
          as a list, set, or or tuple.) The specified keys are emptied from
          the cache. If None, all keys are emptied from the cache.

        :raises RuntimeError:
            When keys contains an unknown fact.
        """
        refresh_keys = None
        if keys is not None:
            if isinstance("str", type(keys)):
                refresh_keys = (keys,)
            else:
                refresh_keys = keys
        if refresh_keys is not None:
            for key in refresh_keys:
                if key in self._callbacks:
                    if key in self._cache:
                        del self._cache[key]
                else:
                    raise RuntimeError(
                        "The %s fact can not be refreshed. %s "
                        "is not a known fact." % (key, key)
                    )
        else:
            self._cache = dict()
        if exception_on_failure or warnings_on_failure:
            self._exception_on_failure = exception_on_failure
            self._warnings_on_failure = warnings_on_failure
            try:
                str(self._device.facts)
            except Exception:
                if exception_on_failure:
                    raise
            finally:
                if warnings_on_failure and self._should_warn:
                    warnings.warn(
                        "Facts gathering is incomplete. "
                        "To know the reason call "
                        '"dev.facts_refresh('
                        'exception_on_failure=True)"',
                        RuntimeWarning,
                    )
                self._exception_on_failure = False
                self._warnings_on_failure = False
                self._should_warn = False

    # In case optimization flag is enabled, it strips of docstring and __doc__ becomes None
    if __doc__ is None:
        __doc__ = ""

    # Precede the class's documentation with the documentation on the specific
    # facts from  the jnpr.junos.facts package.
    __doc__ = facts_doc + "Implementation details on the _FactCache class:" + __doc__
